import os
import math
import itertools
from datetime import datetime

from ..errors import logger
from ..backends.rasterio_ import align_bounds, array_bounds, aligned_target
from .conversion import Converters
from .base import PropertyMixin as _PropertyMixin
from .util import lazy_wombat

import numpy as np
from scipy.spatial import cKDTree
import pandas as pd
import geopandas as gpd
import xarray as xr
import dask
import dask.array as da
from rasterio.crs import CRS
from rasterio import features
from affine import Affine

try:
    import arosics
    AROSICS_INSTALLED = True
except:
    AROSICS_INSTALLED = False

try:
    import pymorph
    PYMORPH_INSTALLED = True
except:
    PYMORPH_INSTALLED = False


def _remove_near_points(dataframe, r):

    """
    Removes points less than a specified distance to another point

    Args:
        dataframe (GeoDataFrame): The ``GeoDataFrame`` with point geometry.
        r (float or int): The minimum distance (radius) in the CRS units of ``dataframe``.

    Returns:
        ``geopandas.GeoDataFrame``
    """

    # Setup a KD tree
    tree = cKDTree(np.c_[dataframe.geometry.x, dataframe.geometry.y])

    # Query all pairs within ``min_dist`` of each other
    near_pairs = tree.query_pairs(r=r, output_type='ndarray')

    if near_pairs.shape[0] > 0:

        # Get a list of pairs to remove
        rm_idx = list(sorted(set(near_pairs[:, 0].tolist())))

        return dataframe.query("index != {}".format(rm_idx))

    return dataframe


class SpatialOperations(_PropertyMixin):

    def sample(self,
               data,
               method='random',
               band=None,
               n=None,
               strata=None,
               spacing=None,
               min_dist=None,
               max_attempts=10,
               **kwargs):

        """
        Generates samples from a raster

        Args:
            data (DataArray): The ``xarray.DataArray`` to extract data from.
            method (Optional[str]): The sampling method. Choices are ['random', 'systematic'].
            band (Optional[int or str]): The band name to extract from. Only required if ``method`` = 'random' and ``strata`` is given.
            n (Optional[int]): The total number of samples. Only required if ``method`` = 'random'.
            strata (Optional[dict]): The strata to sample within. The dictionary key-->value pairs should be {'conditional,value': sample size}.

                E.g.,

                    strata = {'==,1': 0.5, '>=,2': 0.5}
                    ... would sample 50% of total samples within class 1 and 50% of total samples in class >= 2.

                    strata = {'==,1': 10, '>=,2': 20}
                    ... would sample 10 samples within class 1 and 20 samples in class >= 2.

            spacing (Optional[float]): The spacing (in map projection units) when ``method`` = 'systematic'.
            min_dist (Optional[float or int]): A minimum distance allowed between samples. Only applies when ``method`` = 'random'.
            max_attempts (Optional[int]): The maximum numer of attempts to sample points > ``min_dist`` from each other.
            kwargs (Optional[dict]): Keyword arguments passed to ``geowombat.extract``.

        Returns:
            ``geopandas.GeoDataFrame``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Sample 100 points randomly across the image
            >>> with gw.open('image.tif') as src:
            >>>     df = gw.sample(src, n=100)
            >>>
            >>> # Sample points systematically (with 10km spacing) across the image
            >>> with gw.open('image.tif') as src:
            >>>     df = gw.sample(src, method='systematic', spacing=10000.0)
            >>>
            >>> # Sample 50% of 100 in class 1 and 50% in classes >= 2
            >>> strata = {'==,1': 0.5, '>=,2': 0.5}
            >>> with gw.open('image.tif') as src:
            >>>     df = gw.sample(src, band=1, n=100, strata=strata)
            >>>
            >>> # Specify a per-stratum minimum allowed point distance of 1,000 meters
            >>> with gw.open('image.tif') as src:
            >>>     df = gw.sample(src, band=1, n=100, min_dist=1000, strata=strata)
        """

        if method.strip().lower() not in ['random', 'systematic']:
            raise NameError("The method must be 'random' or 'systematic'.")

        if method.strip().lower() == 'systematic':
            if not isinstance(spacing, float):
                if not isinstance(spacing, int):

                    logger.exception("  If the method is 'systematic', the spacing should be provided as a float or integer.")
                    raise TypeError

        if strata and not band and (method.strip().lower() == 'random'):

            logger.exception('  The band name must be provided with random stratified sampling.')
            raise NameError

        df = None

        if not strata:

            if method == 'systematic':

                x_samples = list()
                y_samples = list()

                for i in range(0, data.gw.nrows, int(spacing / data.gw.celly)):
                    for j in range(0, data.gw.ncols, int(spacing / data.gw.cellx)):

                        x_samples.append(j)
                        y_samples.append(i)

                x_samples = np.array(x_samples, dtype='int64')
                y_samples = np.array(y_samples, dtype='int64')

                # Convert the map indices to map coordinates
                x_coords, y_coords = data.gw.meta.affine * (x_samples, y_samples)

                df = gpd.GeoDataFrame(data=range(0, x_coords.shape[0]),
                                      geometry=gpd.points_from_xy(x_coords, y_coords),
                                      crs=data.crs,
                                      columns=['point'])

            else:

                dfs = None
                sample_size = n
                attempts = 0

                while True:

                    if attempts >= max_attempts:

                        logger.warning('  Max attempts reached. Try relaxing the distance threshold.')
                        break

                        # Sample directly from the coordinates
                    y_coords = np.random.choice(data.y.values, size=sample_size, replace=False)
                    x_coords = np.random.choice(data.x.values, size=sample_size, replace=False)

                    if isinstance(dfs, gpd.GeoDataFrame):

                        dfs = pd.concat((dfs, gpd.GeoDataFrame(data=range(0, x_coords.shape[0]),
                                                               geometry=gpd.points_from_xy(x_coords, y_coords),
                                                               crs=data.crs,
                                                               columns=['point'])), axis=0)

                    else:

                        dfs = gpd.GeoDataFrame(data=range(0, x_coords.shape[0]),
                                               geometry=gpd.points_from_xy(x_coords, y_coords),
                                               crs=data.crs,
                                               columns=['point'])

                    if isinstance(min_dist, float) or isinstance(min_dist, int):

                        # Remove samples within a minimum distance
                        dfn = _remove_near_points(dfs, min_dist)

                        df_diff = dfs.shape[0] - dfn.shape[0]

                        if df_diff > 0:
                            dfs = dfn.copy()
                            sample_size = df_diff

                            attempts += 1

                            continue

                    break

                df = dfs.copy()

        else:

            counter = 0
            dfs = None

            for cond, stratum_size in strata.items():

                sign, value = cond.split(',')
                sign = sign.strip()
                value = float(value)

                if isinstance(stratum_size, int):
                    sample_size = stratum_size
                else:
                    sample_size = int(n * stratum_size)

                attempts = 0

                while True:

                    if attempts >= 50:

                        logger.warning('  Max attempts reached for value {:f}. Try relaxing the distance threshold.'.format(value))

                        if not isinstance(df, gpd.GeoDataFrame):
                            df = dfs.copy()
                        else:
                            df = pd.concat((df, dfs), axis=0)

                        break

                    if sign == '>':
                        valid_samples = da.where(data.sel(band=band).data > value)
                    elif sign == '>=':
                        valid_samples = da.where(data.sel(band=band).data >= value)
                    elif sign == '<':
                        valid_samples = da.where(data.sel(band=band).data < value)
                    elif sign == '<=':
                        valid_samples = da.where(data.sel(band=band).data <= value)
                    elif sign == '==':
                        valid_samples = da.where(data.sel(band=band).data == value)
                    else:
                        logger.exception("  The conditional sign was not recognized. Use one of '>', '>=', '<', '<=', or '=='.")
                        raise NameError

                    valid_samples = dask.compute(valid_samples)[0]

                    y_samples = valid_samples[0]
                    x_samples = valid_samples[1]

                    if y_samples.shape[0] > 0:

                        # Get indices within the stratum
                        idx = np.random.choice(range(0, y_samples.shape[0]), size=sample_size, replace=False)

                        y_samples = y_samples[idx]
                        x_samples = x_samples[idx]

                        # Convert the map indices to map coordinates
                        x_coords, y_coords = data.gw.meta.affine * (x_samples, y_samples)

                        if isinstance(dfs, gpd.GeoDataFrame):

                            dfs = pd.concat((dfs, gpd.GeoDataFrame(data=range(0, x_coords.shape[0]),
                                                                   geometry=gpd.points_from_xy(x_coords, y_coords),
                                                                   crs=data.crs,
                                                                   columns=['point'])), axis=0)

                        else:

                            dfs = gpd.GeoDataFrame(data=range(0, x_coords.shape[0]),
                                                   geometry=gpd.points_from_xy(x_coords, y_coords),
                                                   crs=data.crs,
                                                   columns=['point'])

                        if isinstance(min_dist, float) or isinstance(min_dist, int):

                            # Remove samples within a minimum distance
                            dfn = _remove_near_points(dfs, min_dist)

                            df_diff = dfs.shape[0] - dfn.shape[0]

                            if df_diff > 0:
                                dfs = dfn.copy()
                                sample_size = df_diff

                                attempts += 1

                                continue

                        if not isinstance(df, gpd.GeoDataFrame):
                            df = dfs.copy()
                        else:
                            df = pd.concat((df, dfs), axis=0)

                        dfs = None

                    break

                counter += 1

        if isinstance(df, gpd.GeoDataFrame):
            return self.extract(data, df, **kwargs)
        else:
            return None

    def extract(self,
                data,
                aoi,
                bands=None,
                time_names=None,
                band_names=None,
                frac=1.0,
                all_touched=False,
                id_column='id',
                mask=None,
                n_jobs=8,
                verbose=0,
                **kwargs):

        """
        Extracts data within an area or points of interest. Projections do not need to match,
        as they are handled 'on-the-fly'.

        Args:
            data (DataArray): The ``xarray.DataArray`` to extract data from.
            aoi (str or GeoDataFrame): A file or ``geopandas.GeoDataFrame`` to extract data frame.
            bands (Optional[int or 1d array-like]): A band or list of bands to extract.
                If not given, all bands are used. Bands should be GDAL-indexed (i.e., the first band is 1, not 0).
            band_names (Optional[list]): A list of band names. Length should be the same as `bands`.
            time_names (Optional[list]): A list of time names.
            frac (Optional[float]): A fractional subset of points to extract in each polygon feature.
            all_touched (Optional[bool]): The ``all_touched`` argument is passed to ``rasterio.features.rasterize``.
            id_column (Optional[str]): The id column name.
            mask (Optional[GeoDataFrame or Shapely Polygon]): A ``shapely.geometry.Polygon`` mask to subset to.
            n_jobs (Optional[int]): The number of features to rasterize in parallel.
            verbose (Optional[int]): The verbosity level.
            kwargs (Optional[dict]): Keyword arguments passed to ``dask.compute``.

        Returns:
            ``geopandas.GeoDataFrame``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     df = gw.extract(ds, 'poly.gpkg')
        """

        sensor = self.check_sensor(data, return_error=False)

        band_names = self.check_sensor_band_names(data, sensor, band_names)

        converters = Converters()

        shape_len = data.gw.ndims

        if isinstance(bands, list):
            bands_idx = (np.array(bands, dtype='int64') - 1).tolist()
        elif isinstance(bands, np.ndarray):
            bands_idx = (bands - 1).tolist()
        elif isinstance(bands, int):
            bands_idx = [bands]
        else:

            if shape_len > 2:
                bands_idx = slice(0, None)

        df = converters.prepare_points(data,
                                       aoi,
                                       frac=frac,
                                       all_touched=all_touched,
                                       id_column=id_column,
                                       mask=mask,
                                       n_jobs=n_jobs,
                                       verbose=verbose)

        if verbose > 0:
            logger.info('  Extracting data ...')

        # Convert the map coordinates to indices
        x, y = converters.coords_to_indices(df.geometry.x.values,
                                            df.geometry.y.values,
                                            data.transform)

        vidx = (y.tolist(), x.tolist())

        if shape_len > 2:

            vidx = (bands_idx,) + vidx

            if shape_len > 3:

                # The first 3 dimensions are (bands, rows, columns)
                # TODO: allow user-defined time slice?
                for b in range(0, shape_len - 3):
                    vidx = (slice(0, None),) + vidx

        # Get the raster values for each point
        # TODO: allow neighbor indexing
        res = data.data.vindex[vidx].compute(**kwargs)

        if len(res.shape) == 1:
            df[band_names[0]] = res.flatten()
        elif len(res.shape) == 2:

            # `res` is shaped [samples x dimensions]
            df = pd.concat((df, pd.DataFrame(data=res, columns=band_names)), axis=1)

        else:

            # `res` is shaped [samples x time x dimensions]
            if time_names:

                if isinstance(time_names[0], datetime):
                    time_names = list(itertools.chain(*[[t.strftime('%Y-%m-%d')]*res.shape[2] for t in time_names]))
                else:
                    time_names = list(itertools.chain(*[[t]*res.shape[2] for t in time_names]))

            else:
                time_names = list(itertools.chain(*[['t{:d}'.format(t)]*res.shape[2] for t in range(1, res.shape[1]+1)]))

            band_names_concat = ['{}_{}'.format(a, b) for a, b in list(zip(time_names, band_names*res.shape[1]))]

            df = pd.concat((df,
                            pd.DataFrame(data=res.reshape(res.shape[0],
                                                          res.shape[1]*res.shape[2]),
                                         columns=band_names_concat)),
                           axis=1)

        return df

    def clip(self,
             data,
             df,
             query=None,
             mask_data=False,
             expand_by=0):

        """
        Clips a DataArray by vector polygon geometry

        Args:
            data (DataArray): The ``xarray.DataArray`` to subset.
            df (GeoDataFrame or str): The ``geopandas.GeoDataFrame`` or filename to clip to.
            query (Optional[str]): A query to apply to ``df``.
            mask_data (Optional[bool]): Whether to mask values outside of the ``df`` geometry envelope.
            expand_by (Optional[int]): Expand the clip array bounds by ``expand_by`` pixels on each side.

        Returns:
             ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     ds = gw.clip(ds, df, query="Id == 1")
            >>>
            >>> # or
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     ds = ds.gw.clip(df, query="Id == 1")
        """

        if isinstance(df, str) and os.path.isfile(df):
            df = gpd.read_file(df)

        if query:
            df = df.query(query)

        try:

            if data.crs.strip() != CRS.from_dict(df.crs).to_proj4().strip():

                # Re-project the DataFrame to match the image CRS
                df = df.to_crs(data.crs)

        except:

            if data.crs.strip() != CRS.from_proj4(df.crs).to_proj4().strip():
                df = df.to_crs(data.crs)

        row_chunks = data.gw.row_chunks
        col_chunks = data.gw.col_chunks

        left, bottom, right, top = df.total_bounds

        # Align the geometry array grid
        align_transform, align_width, align_height = align_bounds(left,
                                                                  bottom,
                                                                  right,
                                                                  top,
                                                                  data.res)

        # Get the new bounds
        new_left, new_bottom, new_right, new_top = array_bounds(align_height,
                                                                align_width,
                                                                align_transform)

        if expand_by > 0:

            new_left -= data.gw.cellx*expand_by
            new_bottom -= data.gw.celly*expand_by
            new_right += data.gw.cellx*expand_by
            new_top += data.gw.celly*expand_by

        # Subset the array
        data = self.subset(data,
                           left=new_left,
                           bottom=new_bottom,
                           right=new_right,
                           top=new_top)

        if mask_data:

            # Rasterize the geometry and store as a DataArray
            mask = xr.DataArray(data=da.from_array(features.rasterize(list(df.geometry.values),
                                                                      out_shape=(align_height, align_width),
                                                                      transform=align_transform,
                                                                      fill=0,
                                                                      out=None,
                                                                      all_touched=True,
                                                                      default_value=1,
                                                                      dtype='int32'),
                                                   chunks=(row_chunks, col_chunks)),
                                dims=['y', 'x'],
                                coords={'y': data.y.values,
                                        'x': data.x.values})

            # Return the clipped array
            return data.where(mask == 1)

        else:
            return data

    @staticmethod
    @lazy_wombat
    def mask(data,
             dataframe,
             query=None,
             keep='in'):

        """
        Masks a DataArray by vector polygon geometry

        Args:
            data (DataArray): The ``xarray.DataArray`` to mask.
            dataframe (GeoDataFrame or str): The ``geopandas.GeoDataFrame`` or filename to use for masking.
            query (Optional[str]): A query to apply to ``dataframe``.
            keep (Optional[str]): If ``keep`` = 'in', mask values outside of the geometry (keep inside).
                Otherwise, if ``keep`` = 'out', mask values inside (keep outside).

        Returns:
             ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif') as ds:
            >>>     ds = ds.gw.mask(df)
        """

        if isinstance(dataframe, str) and os.path.isfile(dataframe):
            dataframe = gpd.read_file(dataframe)

        if query:
            dataframe = dataframe.query(query)

        try:

            if data.crs.strip() != CRS.from_dict(dataframe.crs).to_proj4().strip():

                # Re-project the DataFrame to match the image CRS
                dataframe = dataframe.to_crs(data.crs)

        except:

            if data.crs.strip() != CRS.from_proj4(dataframe.crs).to_proj4().strip():
                dataframe = dataframe.to_crs(data.crs)

        # Rasterize the geometry and store as a DataArray
        mask = xr.DataArray(data=da.from_array(features.rasterize(list(dataframe.geometry.values),
                                                                  out_shape=(data.gw.nrows, data.gw.ncols),
                                                                  transform=data.transform,
                                                                  fill=0,
                                                                  out=None,
                                                                  all_touched=True,
                                                                  default_value=1,
                                                                  dtype='int32'),
                                               chunks=(data.gw.row_chunks, data.gw.col_chunks)),
                            dims=['y', 'x'],
                            coords={'y': data.y.values,
                                    'x': data.x.values})

        # Return the masked array
        if keep == 'out':
            return data.where(mask != 1)
        else:
            return data.where(mask == 1)

    @staticmethod
    def subset(data,
               left=None,
               top=None,
               right=None,
               bottom=None,
               rows=None,
               cols=None,
               center=False,
               mask_corners=False):

        """
        Subsets a DataArray

        Args:
            data (DataArray): The ``xarray.DataArray`` to subset.
            left (Optional[float]): The left coordinate.
            top (Optional[float]): The top coordinate.
            right (Optional[float]): The right coordinate.
            bottom (Optional[float]): The bottom coordinate.
            rows (Optional[int]): The number of output rows.
            cols (Optional[int]): The number of output rows.
            center (Optional[bool]): Whether to center the subset on ``left`` and ``top``.
            mask_corners (Optional[bool]): Whether to mask corners (*requires ``pymorph``).

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>>
            >>> with gw.open('image.tif', chunks=512) as ds:
            >>>
            >>>     ds_sub = gw.subset(ds,
            >>>                        left=-263529.884,
            >>>                        top=953985.314,
            >>>                        rows=2048,
            >>>                        cols=2048)
        """

        if isinstance(right, int) or isinstance(right, float):
            cols = int((right - left) / data.gw.celly)

        if not isinstance(cols, int):

            logger.exception('  The right coordinate or columns must be specified.')
            raise NameError

        if isinstance(bottom, int) or isinstance(bottom, float):
            rows = int((top - bottom) / data.gw.celly)

        if not isinstance(rows, int):

            logger.exception('  The bottom coordinate or rows must be specified.')
            raise NameError

        x_idx = np.linspace(math.ceil(left), math.ceil(left) + (cols * abs(data.gw.cellx)), cols) + abs(data.gw.cellxh)
        y_idx = np.linspace(math.ceil(top), math.ceil(top) - (rows * abs(data.gw.celly)), rows) - abs(data.gw.cellyh)

        if center:

            y_idx += ((rows / 2.0) * abs(data.gw.celly))
            x_idx -= ((cols / 2.0) * abs(data.gw.cellx))

        ds_sub = data.sel(y=y_idx,
                          x=x_idx,
                          method='nearest')

        if mask_corners:

            if PYMORPH_INSTALLED:

                try:

                    disk = da.from_array(pymorph.sedisk(r=int(rows/2.0))[:rows, :cols],
                                         chunks=ds_sub.data.chunksize).astype('uint8')
                    ds_sub = ds_sub.where(disk == 1)

                except:
                    logger.warning('  Cannot mask corners without a square subset.')

            else:
                logger.warning('  Cannot mask corners without Pymorph.')

        # Update the left and top coordinates
        transform = list(data.transform)

        transform[2] = x_idx[0]
        transform[5] = y_idx[0]

        # Align the coordinates to the target grid
        dst_transform, dst_width, dst_height = aligned_target(Affine(*transform),
                                                              ds_sub.shape[1],
                                                              ds_sub.shape[0],
                                                              data.res)

        ds_sub.attrs['transform'] = dst_transform

        return ds_sub

    @staticmethod
    def coregister(target,
                   reference,
                   **kwargs):

        """
        Co-registers an image, or images, using AROSICS.

        While the required inputs are DataArrays, the intermediate results are stored as NumPy arrays.
        Therefore, memory usage is constrained to the size of the input data. Dask is not used for any of the
        computation in this function.

        Args:
            target (DataArray or str): The target ``xarray.DataArray`` or file name to co-register to ``reference``.
            reference (DataArray or str): The reference ``xarray.DataArray`` or file name used to co-register ``target``.
            kwargs (Optional[dict]): Keyword arguments passed to ``arosics``.

        Reference:
            https://pypi.org/project/arosics

        Returns:
            ``xarray.DataArray``

        Example:
            >>> import geowombat as gw
            >>>
            >>> # Co-register a single image to a reference image
            >>> with gw.open('target.tif') as tar, gw.open('reference.tif') as ref:
            >>>     results = gw.coregister(tar, ref, q=True, ws=(512, 512), max_shift=3, CPUs=4)
            >>>
            >>> # or
            >>>
            >>> results = gw.coregister('target.tif', 'reference.tif', q=True, ws=(512, 512), max_shift=3, CPUs=4)
        """

        import geowombat as gw_

        if not AROSICS_INSTALLED:

            logger.exception('\nAROSICS must be installed to co-register data.\nSee https://pypi.org/project/arosics for details')
            raise NameError

        if isinstance(reference, str):

            if not os.path.isfile(reference):

                logger.exception('  The reference file does not exist.')
                raise OSError

            with gw_.open(reference) as reference:
                pass

        if isinstance(target, str):

            if not os.path.isfile(target):

                logger.exception('  The target file does not exist.')
                raise OSError

            with gw_.open(target) as target:
                pass

        cr = arosics.COREG(reference.filename,
                           target.filename,
                           **kwargs)

        try:
            cr.calculate_spatial_shifts()
        except:

            logger.warning('  Could not co-register the data.')
            return target

        shift_info = cr.correct_shifts()

        left = shift_info['updated geotransform'][0]
        top = shift_info['updated geotransform'][3]

        transform = (target.gw.cellx, 0.0, left, 0.0, -target.gw.celly, top)

        target.attrs['transform'] = transform

        data = shift_info['arr_shifted'].transpose(2, 0, 1)

        ycoords = np.linspace(top-target.gw.cellyh,
                              top-target.gw.cellyh-(data.shape[1] * target.gw.celly),
                              data.shape[1])

        xcoords = np.linspace(left+target.gw.cellxh,
                              left+target.gw.cellxh+(data.shape[2] * target.gw.cellx),
                              data.shape[2])

        return xr.DataArray(data=da.from_array(data,
                                               chunks=(target.gw.band_chunks,
                                                       target.gw.row_chunks,
                                                       target.gw.col_chunks)),
                            dims=('band', 'y', 'x'),
                            coords={'band': target.band.values.tolist(),
                                    'y': ycoords,
                                    'x': xcoords},
                            attrs=target.attrs)



from ..errors import logger

import xarray as xr


def pan_sharpen(data,
                bands=None,
                pan=None,
                blue_weight=1.0,
                green_weight=1.0,
                red_weight=1.0,
                nir_weight=1.0,
                scale_factor=1.0):

    """
    Sharpens wavelengths using the panchromatic band

    Args:
        data (DataArray): The band data.
        bands (Optional[list]): The bands to sharpen. If not given, 'blue', 'green', and 'red' are used.
        pan (Optional[DataArray]): The panchromatic ``DataArray``. ``pan`` is only needed if it is not included
            with ``data``.
        blue_weight (Optional[float]): The blue band weight.
        green_weight (Optional[float]): The green band weight.
        red_weight (Optional[float]): The red band weight.
        nir_weight (Optional[float]): The NIR band weight.
        scale_factor (Optional[float]): A scale factor to apply to the data.

    Example:
        >>> import geowombat as gw
        >>> from geowombat.radiometry import pan_sharpen
        >>>
        >>> with gw.config.update(sensor='l7', scale_factor=0.0001, ref_res=(15, 15)):
        >>>     with gw.open('image.tif', resampling='cubic') as src:
        >>>         pan_sharpen(src)

    Returns:
        ``xarray.DataArray``
    """

    if scale_factor == 1.0:
        scale_factor = data.gw.scale_factor

    if not bands:
        bands = ['blue', 'green', 'red']

    if ','.join(sorted(bands)) != 'blue,green,red':
        if ','.join(sorted(bands)) != 'blue,green,nir,red':
            logger.exception('  The bands must be blue,green,red or blue,green,red,nir')

    attrs = data.attrs.copy()

    data = data * scale_factor

    if isinstance(pan, xr.DataArray):
        pan = pan.sel(band='pan') * scale_factor
    else:
        pan = data.sel(band='pan')

    if ','.join(sorted(bands)) == 'blue,green,red':

        weights = blue_weight + green_weight + red_weight

        band_avg = (data.sel(band='blue') * blue_weight +
                    data.sel(band='green') * green_weight +
                    data.sel(band='red') * red_weight) / weights

    else:

        weights = blue_weight + green_weight + red_weight + nir_weight

        band_avg = (data.sel(band='blue') * blue_weight +
                    data.sel(band='green') * green_weight +
                    data.sel(band='red') * red_weight +
                    data.sel(band='nir') * nir_weight) / weights

    dnf = pan / band_avg

    data_sharp = data * dnf

    data_sharp = data_sharp.assign_coords(coords={'band': bands})

    data_sharp = (data_sharp / scale_factor).astype(data.dtype)

    return data_sharp.assign_attrs(**attrs)



from .angles import landsat_pixel_angles, sentinel_pixel_angles
from .brdf import BRDF
from .topo import Topo
from .sr import LinearAdjustments, RadTransforms
from .qa import QAMasker
from .sharpen import pan_sharpen

__all__ = ['landsat_pixel_angles',
           'sentinel_pixel_angles',
           'BRDF',
           'Topo',
           'LinearAdjustments',
           'RadTransforms',
           'QAMasker',
           'pan_sharpen']



import os
from pathlib import Path
import shutil
import itertools
import fnmatch
import ctypes
import concurrent.futures
from contextlib import contextmanager
import multiprocessing as multi
import threading
import random
import string

from ..errors import logger
from ..backends.rasterio_ import WriteDaskArray
from ..backends.zarr_ import to_zarr
from .windows import get_window_offsets

import numpy as np
import geopandas as gpd
import xarray as xr

import dask.array as da
from dask import is_dask_collection
from dask.diagnostics import ProgressBar
from dask.distributed import Client, LocalCluster

import rasterio as rio
from rasterio.windows import Window
from rasterio.vrt import WarpedVRT
from rasterio.enums import Resampling
from rasterio import shutil as rio_shutil

from affine import Affine

import zarr
from tqdm import tqdm

try:
    MKL_LIB = ctypes.CDLL('libmkl_rt.so')
except:
    MKL_LIB = None


def get_norm_indices(n_bands, window_slice, indexes_multi):

    # Prepend the band position index to the window slice
    if n_bands == 1:

        window_slice = tuple([slice(0, 1)] + list(window_slice))
        indexes = 1

    else:

        window_slice = tuple([slice(0, n_bands)] + list(window_slice))
        indexes = indexes_multi

    return window_slice, indexes


def _window_worker(w):

    """
    Helper to return window slice
    """

    return slice(w.row_off, w.row_off + w.height), slice(w.col_off, w.col_off + w.width)


def _window_worker_time(w, n_bands, tidx, n_time):

    """
    Helper to return window slice
    """

    window_slice = (slice(w.row_off, w.row_off + w.height), slice(w.col_off, w.col_off + w.width))

    # Prepend the band position index to the window slice
    if n_bands == 1:
        window_slice = tuple([slice(tidx, n_time)] + [slice(0, 1)] + list(window_slice))
    else:
        window_slice = tuple([slice(tidx, n_time)] + [slice(0, n_bands)] + list(window_slice))

    return window_slice


@contextmanager
def _cluster_dummy(**kwargs):
    yield None


@contextmanager
def _client_dummy(**kwargs):
    yield None


def _compressor(*args):

    w_, b_, f_, o_ = list(itertools.chain(*args))

    with rio.open(f_, mode='r+', sharing=False) as dst_:

        dst_.write(np.squeeze(b_),
                   window=w_,
                   indexes=o_)


def _block_write_func(*args):

    ofn_, fn_, g_, t_ = list(itertools.chain(*args))

    if t_ == 'zarr':

        group_node = zarr.open(fn_, mode='r')[g_]

        w_ = Window(row_off=group_node.attrs['row_off'],
                    col_off=group_node.attrs['col_off'],
                    height=group_node.attrs['height'],
                    width=group_node.attrs['width'])

        out_data_ = np.squeeze(group_node['data'][:])

    else:

        w_ = Window(row_off=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-4][1:]),
                    col_off=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-3][1:]),
                    height=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-2][1:]),
                    width=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-1][1:]))

        with rio.open(fn_) as src_:
            out_data_ = np.squeeze(src_.read(window=w_))

    out_indexes_ = 1 if len(out_data_.shape) == 2 else list(range(1, out_data_.shape[0]+1))

    with rio.open(ofn_, mode='r+', sharing=False) as dst_:

        dst_.write(out_data_,
                   window=w_,
                   indexes=out_indexes_)


def _block_read_func(fn_, g_, t_):

    """
    Function for block writing with ``concurrent.futures``
    """

    # fn_, g_, t_ = list(itertools.chain(*args))

    if t_ == 'zarr':

        group_node = zarr.open(fn_, mode='r')[g_]

        w_ = Window(row_off=group_node.attrs['row_off'],
                    col_off=group_node.attrs['col_off'],
                    height=group_node.attrs['height'],
                    width=group_node.attrs['width'])

        out_data_ = np.squeeze(group_node['data'][:])

    else:

        w_ = Window(row_off=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-4][1:]),
                    col_off=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-3][1:]),
                    height=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-2][1:]),
                    width=int(os.path.splitext(os.path.basename(fn_))[0].split('_')[-1][1:]))

        out_data_ = np.squeeze(rio.open(fn_).read(window=w_))

    out_indexes_ = 1 if len(out_data_.shape) == 2 else list(range(1, out_data_.shape[0]+1))

    return w_, out_indexes_, out_data_


def _compute_block(block, wid, window_, num_workers):

    """
    Computes a DataArray window block of data

    Args:
        block (DataArray): The ``xarray.DataArray`` to compute.
        wid (int): The window id.
        window_ (namedtuple): The window ``rasterio.windows.Window`` object.
        num_workers (int): The number of parallel workers for ``dask.compute``.

    Returns:
        ``numpy.ndarray``, ``rasterio.windows.Window``, ``int`` | ``list``
    """

    if 'apply' in block.attrs:

        # The geo-transform is needed on the block
        left_, top_ = Affine(*block.transform) * (window_.col_off, window_.row_off)

        attrs = block.attrs.copy()

        # Update the block transform
        attrs['transform'] = Affine(block.gw.cellx, 0.0, left_, 0.0, -block.gw.celly, top_)
        attrs['window_id'] = wid

        block = block.assign_attrs(**attrs)

        if hasattr(block.attrs['apply'], 'wombat_func_'):

            if block.attrs['apply'].wombat_func_:

                # Add the data to the keyword arguments
                block.attrs['apply_kwargs']['data'] = block

                out_data_ = block.attrs['apply'](**block.attrs['apply_kwargs'])

                with threading.Lock():
                    out_data_ = out_data_.data.compute(scheduler='threads', num_workers=num_workers)

        else:

            with threading.Lock():
                out_data_ = block.data.compute(scheduler='threads', num_workers=num_workers)

            if ('apply_args' in block.attrs) and ('apply_kwargs' in block.attrs):
                out_data_ = block.attrs['apply'](out_data_, *block.attrs['apply_args'], **block.attrs['apply_kwargs'])
            elif ('apply_args' in block.attrs) and ('apply_kwargs' not in block.attrs):
                out_data_ = block.attrs['apply'](out_data_, *block.attrs['apply_args'])
            elif ('apply_args' not in block.attrs) and ('apply_kwargs' in block.attrs):
                out_data_ = block.attrs['apply'](out_data_, **block.attrs['apply_kwargs'])
            else:
                out_data_ = block.attrs['apply'](out_data_)

    else:

        with threading.Lock():
            out_data_ = block.data.compute(scheduler='threads', num_workers=num_workers)

    dshape = out_data_.shape

    if len(dshape) > 2:
        out_data_ = np.squeeze(out_data_)

    if len(dshape) == 2:
        indexes_ = 1
    else:
        indexes_ = 1 if dshape[0] == 1 else list(range(1, dshape[0]+1))

    return out_data_, window_, indexes_


def _write_xarray(*args):

    """
    Writes a DataArray to file

    Args:
        args (iterable): A tuple from the window generator.

    Reference:
        https://github.com/dask/dask/issues/3600

    Returns:
        ``str`` | None
    """

    zarr_file = None

    block, filename, wid, block_window, n_threads, separate, chunks, root = list(itertools.chain(*args))

    output, out_window, out_indexes = _compute_block(block, wid, block_window, n_threads)

    if separate:
        zarr_file = to_zarr(filename, output, out_window, chunks, root=root)
    else:

        with threading.Lock():

            with rio.open(filename,
                          mode='r+',
                          sharing=False) as dst_:

                dst_.write(output,
                           window=out_window,
                           indexes=out_indexes)

    return zarr_file


def to_vrt(data,
           filename,
           resampling=None,
           nodata=None,
           init_dest_nodata=True,
           warp_mem_limit=128):

    """
    Writes a file to a VRT file

    Args:
        data (DataArray): The ``xarray.DataArray`` to write.
        filename (str): The output file name to write to.
        resampling (Optional[object]): The resampling algorithm for ``rasterio.vrt.WarpedVRT``. Default is 'nearest'.
        nodata (Optional[float or int]): The 'no data' value for ``rasterio.vrt.WarpedVRT``.
        init_dest_nodata (Optional[bool]): Whether or not to initialize output to ``nodata`` for ``rasterio.vrt.WarpedVRT``.
        warp_mem_limit (Optional[int]): The GDAL memory limit for ``rasterio.vrt.WarpedVRT``.

    Example:
        >>> import geowombat as gw
        >>> from rasterio.enums import Resampling
        >>>
        >>> with gw.config.update(ref_crs=102033):
        >>>
        >>>     with gw.open('image.tif') as ds:
        >>>
        >>>         gw.to_vrt(ds,
        >>>                   'image.vrt',
        >>>                   resampling=Resampling.cubic,
        >>>                   warp_mem_limit=256)
    """

    if not resampling:
        resampling = Resampling.nearest

    # Open the input file on disk
    with rio.open(data.filename) as src:

        with WarpedVRT(src,
                       src_crs=src.crs,                         # the original CRS
                       crs=data.crs,                            # the transformed CRS
                       src_transform=src.transform,             # the original transform
                       transform=data.transform,                # the new transform
                       dtype=data.gw.dtype,
                       resampling=resampling,
                       nodata=nodata,
                       init_dest_nodata=init_dest_nodata,
                       warp_mem_limit=warp_mem_limit) as vrt:

            rio_shutil.copy(vrt, filename, driver='VRT')


def to_raster(data,
              filename,
              readxsize=None,
              readysize=None,
              use_dask_store=False,
              separate=False,
              out_block_type='zarr',
              keep_blocks=False,
              verbose=0,
              overwrite=False,
              gdal_cache=512,
              scheduler='mpool',
              n_jobs=1,
              n_workers=None,
              n_threads=None,
              n_chunks=None,
              overviews=False,
              resampling='nearest',
              use_client=False,
              address=None,
              total_memory=48,
              **kwargs):

    """
    Writes a ``dask`` array to a raster file

    Args:
        data (DataArray): The ``xarray.DataArray`` to write.
        filename (str): The output file name to write to.
        readxsize (Optional[int]): The size of column chunks to read. If not given, ``readxsize`` defaults to Dask
            chunk size.
        readysize (Optional[int]): The size of row chunks to read. If not given, ``readysize`` defaults to Dask
            chunk size.
        separate (Optional[bool]): Whether to write blocks as separate files. Otherwise, write to a single file.
        use_dask_store (Optional[bool]): Whether to use ``dask.array.store`` to save with Dask task graphs.
        out_block_type (Optional[str]): The output block type. Choices are ['gtiff', 'zarr'].
            Only used if ``separate`` = ``True``.
        keep_blocks (Optional[bool]): Whether to keep the blocks stored on disk. Only used if ``separate`` = ``True``.
        verbose (Optional[int]): The verbosity level.
        overwrite (Optional[bool]): Whether to overwrite an existing file.
        gdal_cache (Optional[int]): The ``GDAL`` cache size (in MB).
        scheduler (Optional[str]): The ``concurrent.futures`` scheduler to use. Choices are ['processes', 'threads', 'mpool'].

            mpool: process pool of workers using ``multiprocessing.Pool``
            processes: process pool of workers using ``concurrent.futures``
            threads: thread pool of workers using ``concurrent.futures``

        n_jobs (Optional[int]): The total number of parallel jobs.
        n_workers (Optional[int]): The number of processes.
        n_threads (Optional[int]): The number of threads.
        n_chunks (Optional[int]): The chunk size of windows. If not given, equal to ``n_workers`` x 50.
        overviews (Optional[bool or list]): Whether to build overview layers.
        resampling (Optional[str]): The resampling method for overviews when ``overviews`` is ``True`` or a ``list``.
            Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        use_client (Optional[bool]): Whether to use a ``dask`` client.
        address (Optional[str]): A cluster address to pass to client. Only used when ``use_client`` = ``True``.
        total_memory (Optional[int]): The total memory (in GB) required when ``use_client`` = ``True``.
        kwargs (Optional[dict]): Additional keyword arguments to pass to ``rasterio.write``.

    Returns:
        ``dask.delayed`` object

    Examples:
        >>> import geowombat as gw
        >>>
        >>> # Use 8 parallel workers
        >>> with gw.open('input.tif') as ds:
        >>>     gw.to_raster(ds, 'output.tif', n_jobs=8)
        >>>
        >>> # Use 4 process workers and 2 thread workers
        >>> with gw.open('input.tif') as ds:
        >>>     gw.to_raster(ds, 'output.tif', n_workers=4, n_threads=2)
        >>>
        >>> # Control the window chunks passed to concurrent.futures
        >>> with gw.open('input.tif') as ds:
        >>>     gw.to_raster(ds, 'output.tif', n_workers=4, n_threads=2, n_chunks=16)
        >>>
        >>> # Compress the output and build overviews
        >>> with gw.open('input.tif') as ds:
        >>>     gw.to_raster(ds, 'output.tif', n_jobs=8, overviews=True, compress='lzw')
    """

    if MKL_LIB:
        __ = MKL_LIB.MKL_Set_Num_Threads(n_threads)

    pfile = Path(filename)

    if scheduler.lower() == 'mpool':
        pool_executor = multi.Pool
    else:
        pool_executor = concurrent.futures.ProcessPoolExecutor if scheduler.lower() == 'processes' else concurrent.futures.ThreadPoolExecutor

    if overwrite:

        if pfile.is_file():
            pfile.unlink()

    if pfile.is_file():
        logger.warning('  The output file already exists.')
        return

    if not is_dask_collection(data.data):
        logger.exception('  The data should be a dask array.')

    if use_client:

        if address:
            cluster_object = _cluster_dummy
        else:
            cluster_object = LocalCluster

        client_object = Client

    else:

        cluster_object = _cluster_dummy
        client_object = _client_dummy

    if isinstance(n_workers, int) and isinstance(n_threads, int):
        n_jobs = n_workers * n_threads
    else:

        n_workers = n_jobs
        n_threads = 1

    mem_per_core = int(total_memory / n_workers)

    if not isinstance(n_chunks, int):
        n_chunks = n_workers * 50

    if not isinstance(readxsize, int):
        readxsize = data.gw.col_chunks

    if not isinstance(readysize, int):
        readysize = data.gw.row_chunks

    chunksize = (data.gw.row_chunks, data.gw.col_chunks)

    # Force tiled outputs with no file sharing
    kwargs['sharing'] = False

    if data.gw.tiled:
        kwargs['tiled'] = True

    if 'compress' in kwargs:

        # Store the compression type because
        #   it is removed in concurrent writing
        compress = True
        compress_type = kwargs['compress']
        del kwargs['compress']

    elif isinstance(data.gw.compress, str) and data.gw.compress.lower() in ['lzw', 'deflate']:

        compress = True
        compress_type = data.gw.compress

    else:
        compress = False

    if 'nodata' not in kwargs:

        if isinstance(data.gw.nodata, int) or isinstance(data.gw.nodata, float):
            kwargs['nodata'] = data.gw.nodata

    if 'blockxsize' not in kwargs:
        kwargs['blockxsize'] = data.gw.col_chunks

    if 'blockysize' not in kwargs:
        kwargs['blockysize'] = data.gw.row_chunks

    if 'bigtiff' not in kwargs:
        kwargs['bigtiff'] = data.gw.bigtiff

    if 'driver' not in kwargs:
        kwargs['driver'] = data.gw.driver

    if 'count' not in kwargs:
        kwargs['count'] = data.gw.nbands

    if 'width' not in kwargs:
        kwargs['width'] = data.gw.ncols

    if 'height' not in kwargs:
        kwargs['height'] = data.gw.nrows

    if separate:

        d_name = pfile.parent
        sub_dir = d_name.joinpath('sub_tmp_')
        zarr_file = sub_dir.joinpath('data.zarr').as_posix()

        sub_dir.mkdir(parents=True, exist_ok=True)

        root = zarr.open(zarr_file, mode='w')

    else:

        root = None

        if verbose > 0:
            logger.info('  Creating the file ...\n')

        with rio.open(filename, mode='w', **kwargs) as rio_dst:
            pass

    if verbose > 0:
        logger.info('  Writing data to file ...\n')

    with rio.Env(GDAL_CACHEMAX=gdal_cache):

        if not use_dask_store:

            windows = get_window_offsets(data.gw.nrows,
                                         data.gw.ncols,
                                         readysize,
                                         readxsize,
                                         return_as='list')

            n_windows = len(windows)

            # Iterate over the windows in chunks
            for wchunk in range(0, n_windows, n_chunks):

                window_slice = windows[wchunk:wchunk+n_chunks]
                n_windows_slice = len(window_slice)

                if verbose > 0:

                    logger.info('  Windows {:,d}--{:,d} of {:,d} ...'.format(wchunk+1,
                                                                             wchunk+n_windows_slice,
                                                                             n_windows))

                if len(data.shape) == 2:

                    data_gen = ((data[w.row_off:w.row_off + w.height, w.col_off:w.col_off + w.width],
                                 filename, widx, w, n_threads, separate, chunksize, root) for widx, w in enumerate(window_slice))

                elif len(data.shape) == 3:

                    data_gen = ((data[:, w.row_off:w.row_off + w.height, w.col_off:w.col_off + w.width],
                                 filename, widx, w, n_threads, separate, chunksize, root) for widx, w in enumerate(window_slice))

                else:

                    data_gen = ((data[:, :, w.row_off:w.row_off + w.height, w.col_off:w.col_off + w.width],
                                 filename, widx, w, n_threads, separate, chunksize, root) for widx, w in enumerate(window_slice))

                with pool_executor(n_workers) as executor:

                    if scheduler == 'mpool':

                        for zarr_file in tqdm(executor.imap_unordered(_write_xarray,
                                                                      data_gen),
                                              total=n_windows_slice):
                            pass

                    else:

                        for zarr_file in tqdm(executor.map(_write_xarray, data_gen), total=n_windows_slice):
                            pass

            # if overviews:
            #
            #     if not isinstance(overviews, list):
            #         overviews = [2, 4, 8, 16]
            #
            #     if resampling not in ['average', 'bilinear', 'cubic', 'cubic_spline',
            #                           'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest']:
            #
            #         logger.warning("  The resampling method is not supported by rasterio. Setting to 'nearest'")
            #
            #         resampling = 'nearest'
            #
            #     if verbose > 0:
            #         logger.info('  Building pyramid overviews ...')
            #
            #     rio_dst.build_overviews(overviews, getattr(Resampling, resampling))
            #     rio_dst.update_tags(ns='overviews', resampling=resampling)

        else:

            with cluster_object(n_workers=n_workers,
                                threads_per_worker=n_threads,
                                scheduler_port=0,
                                processes=False,
                                memory_limit='{:d}GB'.format(mem_per_core)) as cluster:

                cluster_address = address if address else cluster

                with client_object(address=cluster_address) as client:

                    with WriteDaskArray(filename,
                                        overwrite=overwrite,
                                        separate=separate,
                                        out_block_type=out_block_type,
                                        keep_blocks=keep_blocks,
                                        gdal_cache=gdal_cache,
                                        **kwargs) as dst:

                        # Store the data and return a lazy evaluator
                        res = da.store(da.squeeze(data.data),
                                       dst,
                                       lock=False,
                                       compute=False)

                        if verbose > 0:
                            logger.info('  Writing data to file ...')

                        # Send the data to file
                        #
                        # *Note that the progress bar will
                        #   not work with a client.
                        if use_client:
                            res.compute(num_workers=n_jobs)
                        else:

                            with ProgressBar():
                                res.compute(num_workers=n_jobs)

                        if verbose > 0:
                            logger.info('  Finished writing data to file.')

                        out_block_type = dst.out_block_type
                        keep_blocks = dst.keep_blocks
                        zarr_file = dst.zarr_file
                        sub_dir = dst.sub_dir

        if compress:

            if verbose > 0:
                logger.info('  Compressing output file ...')

            if separate:

                group_keys = list(root.group_keys())
                n_groups = len(group_keys   )

                if out_block_type.lower() == 'zarr':
                    # root = zarr.open(zarr_file, mode='r')
                    open_file = zarr_file
                else:

                    outfiles = sorted(fnmatch.filter(os.listdir(sub_dir), '*.tif'))
                    outfiles = [os.path.join(sub_dir, fn) for fn in outfiles]

                    # data_gen = ((fn, None, 'gtiff') for fn in outfiles)

                kwargs['compress'] = compress_type

                n_windows = len(group_keys)

                # Compress into one file
                with rio.open(filename, mode='w', **kwargs) as dst_:

                    # Iterate over the windows in chunks
                    for wchunk in range(0, n_groups, n_chunks):

                        group_keys_slice = group_keys[wchunk:wchunk + n_chunks]
                        n_windows_slice = len(group_keys_slice)

                        if verbose > 0:

                            logger.info('  Windows {:,d}--{:,d} of {:,d} ...'.format(wchunk + 1,
                                                                                     wchunk + n_windows_slice,
                                                                                     n_windows))

                        ################################################
                        data_gen = ((open_file, group, 'zarr') for group in group_keys_slice)

                        # for f in tqdm(executor.map(_compressor, data_gen), total=n_windows_slice):
                        #     pass
                        #
                        # futures = [executor.submit(_compress_dummy, iter_[0], iter_[1], None) for iter_ in data_gen]
                        #
                        # for f in tqdm(concurrent.futures.as_completed(futures), total=n_windows_slice):
                        #
                        #     out_window, out_block = f.result()
                        #
                        #     dst_.write(np.squeeze(out_block),
                        #                window=out_window,
                        #                indexes=out_indexes_)
                        ################################################

                        # data_gen = ((root, group, 'zarr') for group in group_keys_slice)

                        # for f, g, t in tqdm(data_gen, total=n_windows_slice):
                        #
                        #     out_window, out_indexes, out_block = _block_read_func(f, g, t)

                        # executor.map(_block_write_func, data_gen)

                        with concurrent.futures.ProcessPoolExecutor(max_workers=n_workers) as executor:

                            # Submit all of the tasks as futures
                            futures = [executor.submit(_block_read_func, f, g, t) for f, g, t in data_gen]

                            for f in tqdm(concurrent.futures.as_completed(futures), total=n_windows_slice):

                                out_window, out_indexes, out_block = f.result()

                                dst_.write(out_block,
                                           window=out_window,
                                           indexes=out_indexes)

                        futures = None

                if not keep_blocks:
                    shutil.rmtree(sub_dir)

            else:

                p = Path(filename)

                d_name = p.parent
                f_base, f_ext = os.path.splitext(p.name)

                ld = string.ascii_letters + string.digits
                rstr = ''.join(random.choice(ld) for i in range(0, 9))

                temp_file = d_name.joinpath('{}_temp_{}{}'.format(f_base, rstr, f_ext))

                compress_raster(filename,
                                temp_file.as_posix(),
                                n_jobs=n_jobs,
                                gdal_cache=gdal_cache,
                                compress=compress_type)

                temp_file.rename(filename)

            if verbose > 0:
                logger.info('  Finished compressing')

    if verbose > 0:
        logger.info('\nFinished writing the data.')


def _arg_gen(arg_, iter_):
    for i_ in iter_:
        yield arg_


def apply(infile,
          outfile,
          block_func,
          args=None,
          count=1,
          scheduler='processes',
          gdal_cache=512,
          n_jobs=4,
          overwrite=False,
          **kwargs):

    """
    Applies a function and writes results to file

    Args:
        infile (str): The input file to process.
        outfile (str): The output file.
        block_func (func): The user function to apply to each block. The function should always return the window,
            the data, and at least one argument. The block data inside the function will be a 2d array if the
            input image has 1 band, otherwise a 3d array.
        args (Optional[tuple]): Additional arguments to pass to ``block_func``.
        count (Optional[int]): The band count for the output file.
        scheduler (Optional[str]): The ``concurrent.futures`` scheduler to use. Choices are ['threads', 'processes'].
        gdal_cache (Optional[int]): The ``GDAL`` cache size (in MB).
        n_jobs (Optional[int]): The number of blocks to process in parallel.
        overwrite (Optional[bool]): Whether to overwrite an existing output file.
        kwargs (Optional[dict]): Additional keyword arguments to pass to ``rasterio.open``.

    Returns:
        None

    Examples:
        >>> import geowombat as gw
        >>>
        >>> # Here is a function with no arguments
        >>> def my_func0(w, block, arg):
        >>>     return w, block
        >>>
        >>> gw.apply('input.tif',
        >>>          'output.tif',
        >>>           my_func0,
        >>>           n_jobs=8)
        >>>
        >>> # Here is a function with 1 argument
        >>> def my_func1(w, block, arg):
        >>>     return w, block * arg
        >>>
        >>> gw.apply('input.tif',
        >>>          'output.tif',
        >>>           my_func1,
        >>>           args=(10.0,),
        >>>           n_jobs=8)
    """

    if not args:
        args = (None,)

    if overwrite:

        if os.path.isfile(outfile):
            os.remove(outfile)

    kwargs['sharing'] = False
    kwargs['tiled'] = True

    io_mode = 'r+' if os.path.isfile(outfile) else 'w'

    out_indexes = 1 if count == 1 else list(range(1, count+1))

    futures_executor = concurrent.futures.ThreadPoolExecutor if scheduler == 'threads' else concurrent.futures.ProcessPoolExecutor

    with rio.Env(GDAL_CACHEMAX=gdal_cache):

        with rio.open(infile) as src:

            profile = src.profile.copy()

            if 'dtype' not in kwargs:
                kwargs['dtype'] = profile['dtype']

            if 'nodata' not in kwargs:
                kwargs['nodata'] = profile['nodata']

            if 'blockxsize' not in kwargs:
                kwargs['blockxsize'] = profile['blockxsize']

            if 'blockxsize' not in kwargs:
                kwargs['blockysize'] = profile['blockysize']

            # Create a destination dataset based on source params. The
            # destination will be tiled, and we'll process the tiles
            # concurrently.
            profile.update(count=count,
                           **kwargs)

            with rio.open(outfile, io_mode, **profile) as dst:

                # Materialize a list of destination block windows
                # that we will use in several statements below.
                # windows = get_window_offsets(src.height,
                #                              src.width,
                #                              blockysize,
                #                              blockxsize, return_as='list')

                # This generator comprehension gives us raster data
                # arrays for each window. Later we will zip a mapping
                # of it with the windows list to get (window, result)
                # pairs.
                # if nbands == 1:
                data_gen = (src.read(window=w, out_dtype=profile['dtype']) for ij, w in src.block_windows(1))
                # else:
                #
                #     data_gen = (np.array([rio.open(fn).read(window=w,
                #                                             out_dtype=dtype) for fn in infile], dtype=dtype)
                #                 for ij, w in src.block_windows(1))

                if args:
                    args = [_arg_gen(arg, src.block_windows(1)) for arg in args]

                with futures_executor(max_workers=n_jobs) as executor:

                    # Submit all of the tasks as futures
                    futures = [executor.submit(block_func,
                                               iter_[0][1],    # window object
                                               *iter_[1:])     # other arguments
                               for iter_ in zip(list(src.block_windows(1)), data_gen, *args)]

                    for f in tqdm(concurrent.futures.as_completed(futures), total=len(futures)):

                        out_window, out_block = f.result()

                        dst.write(np.squeeze(out_block),
                                  window=out_window,
                                  indexes=out_indexes)

                    # We map the block_func() function over the raster
                    # data generator, zip the resulting iterator with
                    # the windows list, and as pairs come back we
                    # write data to the destination dataset.
                    # for window_tuple, result in tqdm(zip(list(src.block_windows(1)),
                    #                                      executor.map(block_func,
                    #                                                   data_gen,
                    #                                                   *args)),
                    #                                  total=n_windows):
                    #
                    #     dst.write(result,
                    #               window=window_tuple[1],
                    #               indexes=out_indexes)


def _compress_dummy(w, block, dummy):

    """
    Dummy function to pass to concurrent writing
    """

    return w, block


def compress_raster(infile, outfile, n_jobs=1, gdal_cache=512, compress='lzw'):

    """
    Compresses a raster file

    Args:
        infile (str): The file to compress.
        outfile (str): The output file.
        n_jobs (Optional[int]): The number of concurrent blocks to write.
        gdal_cache (Optional[int]): The ``GDAL`` cache size (in MB).
        compress (Optional[str]): The compression method.

    Returns:
        None
    """

    with rio.open(infile) as src:

        profile = src.profile.copy()

        profile.update(compress=compress)

        apply(infile,
              outfile,
              _compress_dummy,
              scheduler='processes',
              args=(None,),
              gdal_cache=gdal_cache,
              n_jobs=n_jobs,
              count=src.count,
              dtype=src.profile['dtype'],
              nodata=src.profile['nodata'],
              tiled=src.profile['tiled'],
              blockxsize=src.profile['blockxsize'],
              blockysize=src.profile['blockysize'],
              compress=compress)



import xarray as xr
import dask.array as da


class QAMasker(object):

    """
    A class for masking bit-packed quality flags

    Args:
        qa (DataArray): The band quality array.
        sensor (str): The sensor name. Choices are ['hls', 'l8-pre', 'l8-c1', 'l-c1', 'modis'].
                'l-c1' refers to Collection 1 L4-5 and L7. 'l8-c1' refers to Collection 1 L8.
        mask_items (str list): A list of items to mask.
        modis_qa_position (Optional[int]): The MODIS QA band position. Default is 1.
        modis_quality (Optional[int]): The MODIS quality level. Default is 2.
        confidence_level (Optional[str]): The confidence level. Choices are ['notdet', 'no', 'maybe', 'yes'].

    References:
        Landsat Collection 1:
            https://landsat.usgs.gov/collectionqualityband

    Examples:
        >>> import geowombat as gw
        >>> from geowombat.radiometry import QAMasker
        >>>
        >>> # Get the MODIS cloud mask.
        >>> with gw.open('qa.tif') as qa:
        >>>     mask = QAMasker(qs, 'modis').to_mask()
        >>>
        >>> # NASA HLS
        >>> with gw.open('qa.tif') as qa:
        >>>     mask = QAMasker(qs, 'hls', ['cloud']).to_mask()
    """

    def __init__(self,
                 qa,
                 sensor,
                 mask_items=None,
                 modis_qa_band=1,
                 modis_quality=2,
                 confidence_level='yes'):

        self.qa = qa
        self.sensor = sensor
        self.modis_qa_band = modis_qa_band
        self.modis_quality = modis_quality
        self.mask_items = mask_items
        self.confidence_level = confidence_level

        self._set_dicts()

    def to_mask(self):

        """
        Converts QA bit-packed data to an integer mask

        Returns:

            ``xarray.DataArray``:

                0: clear,
                1: water,
                2: shadow,
                3: snow or ice,
                4: cloud,
                5: cirrus cloud,
                6: adjacent cloud,
                7: saturated,
                8: dropped,
                9: terrain occluded,
                255: fill
        """

        if self.sensor == 'MODIS':
            mask = self._get_modis_qa_mask()
        else:

            mask = da.zeros((self.qa.gw.nrows, self.qa.gw.ncols),
                            chunks=(self.qa.gw.row_chunks, self.qa.gw.col_chunks),
                            dtype='uint8')

            for mask_item in self.mask_items:

                if mask_item in self.qa_flags[self.sensor]:

                    if 'conf' in mask_item:

                        # Has high confidence that
                        #   this condition was met.
                        mask_value = self.conf_dict[self.confidence_level]

                    else:
                        mask_value = 1

                    mask = da.where(self._get_qa_mask(mask_item) >= mask_value, self.fmask_dict[mask_item], mask)

        mask = xr.DataArray(mask,
                            dims=('y', 'x'),
                            coords={'y': self.qa.y,
                                    'x': self.qa.x},
                            attrs=self.qa.attrs)

        mask = mask.expand_dims(dim='band')
        mask = mask.assign_coords(band=['mask'])

        for k, v in self.fmask_dict.items():
            mask.attrs[k] = v

        return mask

    def _set_dicts(self):

        self.fmask_dict = dict(clear=0,
                               water=1,
                               shadow=2,
                               shadowconf=2,
                               snow=3,
                               snowice=3,
                               snowiceconf=3,
                               cloud=4,
                               cloudconf=4,
                               cirrus=5,
                               cirrusconf=5,
                               adjacent=6,
                               saturated=7,
                               dropped=8,
                               terrain=0,
                               fill=255)

        self.conf_dict = dict(notdet=0,
                              no=1,
                              maybe=2,
                              yes=3)

        self.qa_flags = {'hls': {'cirrus': (0, 0),
                                 'cloud': (1, 1),
                                 'adjacent': (2, 2),
                                 'shadow': (3, 3),
                                 'snowice': (4, 4),
                                 'water': (5, 5)},
                         'l8-pre': {'cirrus': (13, 12),
                                    'snowice': (11, 10),
                                    'water': (5, 4),
                                    'fill': (0, 0),
                                    'dropped': (1, 1),
                                    'terrain': (2, 2),
                                    'shadow': (7, 6),
                                    'vegconf': (9, 8),
                                    'snowiceconf': (11, 10),
                                    'cirrusconf': (13, 12),
                                    'cloudconf': (15, 14)},
                         'l8-c1': {'cirrusconf': (12, 11),
                                   'snowiceconf': (10, 9),
                                   'shadowconf': (8, 7),
                                   'cloudconf': (6, 5),
                                   'cloud': (4, 4),
                                   'saturated': (3, 2),
                                   'terrain': (1, 1),
                                   'fill': (0, 0)},
                         'l-c1': {'fill': (0, 0),
                                  'dropped': (1, 1),
                                  'saturated': (3, 2),
                                  'cloud': (4, 4),
                                  'cloudconf': (6, 5),
                                  'shadowconf': (8, 7),
                                  'snowice': (10, 9)},
                         'ard': {'fill': (0, 0),
                                 'clear': (1, 1),
                                 'water': (2, 2),
                                 'shadow': (3, 3),
                                 'snow': (4, 4),
                                 'cloud': (5, 5)},
                         'modis': {'cloud': (0, 0),
                                   'daynight': (3, 3),
                                   'sunglint': (4, 4),
                                   'snowice': (5, 5),
                                   'landwater': (7, 6)}}

        self.modis_bit_shifts = {1: 0,
                                 2: 4,
                                 3: 8,
                                 4: 12,
                                 5: 16,
                                 6: 20,
                                 7: 24}

    def _qa_bits(self, mask_item):

        """
        Args:
            mask_item (str)

        For confidence bits:
            0 = not determined
            1 = no
            2 = maybe
            3 = yes
        """

        bit_location = self.qa_flags[self.sensor][mask_item]

        self.b1 = bit_location[0]
        self.b2 = bit_location[1]

    def _get_modis_qa_mask(self):

        """
        Reference:
            https://github.com/haoliangyu/pymasker/blob/master/pymasker.py
        """

        # `modis_mask`
        #   0: best quality
        #   1: good quality
        #   4: fill value
        #
        # `output`
        #   0: good data = clear
        #   255: bad data = fill
        return np.where(np.uint8(self.qa >> self.modis_bit_shifts[self.modis_qa_band] & 4) <= self.modis_quality,
                        self.fmask_dict['clear'],
                        self.fmask_dict['fill'])

    def _get_qa_mask(self, mask_item):

        """
        Args:
            mask_item (str)

        Reference:
            https://github.com/mapbox/landsat8-qa/blob/master/landsat8_qa/qa.py
        """

        self._qa_bits(mask_item)

        width_int = int((self.b1 - self.b2 + 1) * '1', 2)

        return ((self.qa.data.squeeze() >> self.b2) & width_int).astype('uint8')



import os
import shutil
import fnmatch
import tarfile
import subprocess
from pathlib import Path
from datetime import datetime
from collections import namedtuple
import random
import string

from ..errors import logger
from ..radiometry import BRDF, LinearAdjustments, RadTransforms, landsat_pixel_angles, sentinel_pixel_angles, QAMasker

import geowombat as gw

import numpy as np
import pandas as pd
import geopandas as gpd
import xarray as xr
import shapely
from shapely.geometry import Polygon
# import wget


shapely.speedups.enable()


def _random_id(string_length):

    """
    Generates a random string of letters and digits
    """

    letters_digits = string.ascii_letters + string.digits

    return ''.join(random.choice(letters_digits) for i in range(string_length))


def _parse_google_filename(filename, landsat_parts, sentinel_parts, public_url):

    FileInfo = namedtuple('FileInfo', 'url url_file meta angles')

    file_info = FileInfo(url=None, url_file=None, meta=None, angles=None)

    f_base, f_ext = os.path.splitext(filename)

    fn_parts = f_base.split('_')

    if fn_parts[0].lower() in landsat_parts:

        # Collection 1
        url_ = '{PUBLIC}-landsat/{SENSOR}/01/{PATH}/{ROW}/{FDIR}'.format(PUBLIC=public_url,
                                                                         SENSOR=fn_parts[0],
                                                                         PATH=fn_parts[2][:3],
                                                                         ROW=fn_parts[2][3:],
                                                                         FDIR='_'.join(fn_parts[:-1]))

        url_filename = '{URL}/{FN}'.format(URL=url_, FN=filename)
        url_meta = '{URL}/{FN}_MTL.txt'.format(URL=url_, FN='_'.join(fn_parts[:-1]))
        url_angles = '{URL}/{FN}_ANG.txt'.format(URL=url_, FN='_'.join(fn_parts[:-1]))

        file_info = FileInfo(url=url_,
                             url_file=url_filename,
                             meta=url_meta,
                             angles=url_angles)

    # elif fn_parts[0].lower() in sentinel_parts:
    #
    #     safe_dir = '{SENSOR}_{M}{L}_'.format(SENSOR=fn_parts[0], M=fn_parts[2], L=fn_parts[3])
    #
    #     '/01/K/AB/S2A_MSIL1C_20160519T222025_N0202_R029_T01KAB_20160520T024950.SAFE/GRANULE/S2A_OPER_MSI_L1C_TL_SGS__20160519T234326_A004745_T01KAB_N02.02/IMG_DATA/S2A_OPER_MSI_L1C_TL_SGS__20160519T234326_A004745_T01KAB_B05.jp2'
    #
    #     fn = '{PUBLIC}-sentinel-2/tiles/{UTM}/{LAT}/{GRID}'.format(PUBLIC=public_url,
    #                                                                UTM=,
    #                                                                LAT=,
    #                                                                GRID=)

    return file_info


class GeoDownloads(object):

    def __init__(self):

        self.gcp_public = 'https://storage.googleapis.com/gcp-public-data'

        self.landsat_parts = ['lt05', 'le07', 'lc08']
        self.sentinel_parts = ['s2a']

        self.associations = dict(l7=dict(blue=1,
                                         green=2,
                                         red=3,
                                         nir=4,
                                         swir1=5,
                                         thermal=6,
                                         swir2=7,
                                         pan=8),
                                 l8=dict(coastal=1,
                                         blue=2,
                                         green=3,
                                         red=4,
                                         nir=5,
                                         swir1=6,
                                         swir2=7,
                                         pan=8,
                                         cirrus=9,
                                         tirs1=10,
                                         tirs2=11),
                                 s2=dict(coastal=1,
                                         blue=2,
                                         green=3,
                                         red=4,
                                         nir1=5,
                                         nir2=6,
                                         nir3=7,
                                         nir=8,
                                         rededge=8,
                                         water=9,
                                         cirrus=10,
                                         swir1=11,
                                         swir2=12))

        self.search_dict = dict()

    def download_cube(self,
                      sensors,
                      date_range,
                      bounds,
                      bands,
                      crs=None,
                      out_bounds=None,
                      outdir='.',
                      ref_res=None,
                      l57_angles_path=None,
                      l8_angles_path=None,
                      write_angle_files=False,
                      **kwargs):

        """
        Downloads a cube of Landsat and/or Sentinel 2 imagery

        Args:
            sensors (str or list): The sensors, or sensor, to download.
            date_range (list): The date range, given as [date1, date2], where the date format is yyyy-mm-dd.
            bounds (GeoDataFrame, list, or tuple): The geometry bounds (in WGS84 lat/lon) that define the cube extent
                to download. If given as a ``GeoDataFrame``, only the first ``DataFrame`` record will be used.
                If given as a ``tuple`` or a ``list``, the order should be (left, bottom, right, top).
            bands (str or list): The bands to download.
            crs (Optional[str or object]): The output CRS. If ``bounds`` is a ``GeoDataFrame``, the CRS is taken
                from the object.
            out_bounds (Optional[list or tuple]): The output bounds in ``crs``. If not given, the bounds are
                taken from ``bounds``.
            outdir (Optional[str]): The output directory.
            ref_res (Optional[tuple]): A reference cell resolution.
            l57_angles_path (str): The path to the Landsat 5 and 7 angles bin.
            l8_angles_path (str): The path to the Landsat 8 angles bin.
            write_angle_files (Optional[bool]): Whether to write the angles to file.
            kwargs (Optional[dict]): Keyword arguments passed to ``to_raster``.

        Examples:
            >>> from geowombat.util import GeoDownloads
            >>> gdl = GeoDownloads()
            >>>
            >>> # Download a Landsat 7 panchromatic cube
            >>> gdl.download_cube(['l7'],
            >>>                   ['2010-01-01', '2010-02-01'],
            >>>                   (-91.57, 40.37, -91.46, 40.42),
            >>>                   ['pan'],
            >>>                   crs="+proj=aea +lat_1=-5 +lat_2=-42 +lat_0=-32 +lon_0=-60 +x_0=0 +y_0=0 +ellps=aust_SA +units=m +no_defs")
            >>>
            >>> # Download a Landsat 7, 8 and Sentinel 2 cube of the visible spectrum
            >>> gdl.download_cube(['l7', 'l8', 's2'],
            >>>                   ['2017-01-01', '2018-01-01'],
            >>>                   (-91.57, 40.37, -91.46, 40.42),
            >>>                   ['blue', 'green', 'red'],
            >>>                   crs={'init': 'epsg:102033'},
            >>>                   readxsize=1024,
            >>>                   readysize=1024,
            >>>                   n_workers=1,
            >>>                   n_threads=8)
        """

        # TODO: parameterize
        # kwargs = dict(readxsize=1024,
        #               readysize=1024,
        #               verbose=1,
        #               separate=False,
        #               n_workers=1,
        #               n_threads=8,
        #               n_chunks=100,
        #               overwrite=False)

        angle_infos = dict()

        rt = RadTransforms()
        br = BRDF()
        la = LinearAdjustments()

        main_path = Path(outdir)
        outdir_brdf = main_path.joinpath('brdf')

        if not main_path.is_dir():
            main_path.mkdir()

        if not outdir_brdf.is_dir():
            outdir_brdf.mkdir()

        if isinstance(sensors, str):
            sensors = [sensors]

        status = Path(outdir).joinpath('status.txt')

        if not status.is_file():

            with open(status.as_posix(), mode='w') as tx:
                pass

        # Get bounds from geometry
        if isinstance(bounds, tuple) or isinstance(bounds, list):

            bounds = Polygon([(bounds[0], bounds[3]),   # upper left
                              (bounds[2], bounds[3]),   # upper right
                              (bounds[2], bounds[1]),   # lower right
                              (bounds[0], bounds[1]),   # lower left
                              (bounds[0], bounds[3])])  # upper left

            bounds = gpd.GeoDataFrame([0],
                                      geometry=[bounds],
                                      crs={'init': 'epsg:4326'})

        bounds_object = bounds.geometry.values[0]

        if not out_bounds:

            # Project the bounds
            out_bounds = bounds.to_crs(crs).bounds.values[0].tolist()

        # Get WRS file
        data_bin = os.path.realpath(os.path.dirname(__file__))

        data_dir = Path(data_bin).joinpath('../data')

        shp_dict = dict()

        if ('l7' in sensors) or ('l8' in sensors):

            path_tar = Path(data_dir).joinpath('wrs2.tar.gz')
            path_shp = Path(data_dir).joinpath('wrs2_descending.shp')
            wrs = os.path.realpath(path_shp.as_posix())

            if not path_shp.is_file():

                with tarfile.open(os.path.realpath(path_tar.as_posix()), mode='r:gz') as tf:
                    tf.extractall(data_dir.as_posix())

            df_wrs = gpd.read_file(wrs)
            df_wrs = df_wrs[df_wrs.geometry.intersects(bounds_object)]

            if df_wrs.empty:
                logger.warning('  The geometry bounds is empty.')
                return

            shp_dict['wrs'] = df_wrs

        if 's2' in sensors:

            path_tar = Path(data_dir).joinpath('mgrs.tar.gz')
            path_shp = Path(data_dir).joinpath('sentinel2_grid.shp')
            mgrs = os.path.realpath(path_shp.as_posix())

            if not path_shp.is_file():

                with tarfile.open(os.path.realpath(path_tar.as_posix()), mode='r:gz') as tf:
                    tf.extractall(data_dir.as_posix())

            df_mgrs = gpd.read_file(mgrs)
            df_mgrs = df_mgrs[df_mgrs.geometry.intersects(bounds_object)]

            if df_mgrs.empty:
                logger.warning('  The geometry bounds is empty.')
                return

            shp_dict['mgrs'] = df_mgrs

        dt1 = datetime.strptime(date_range[0], '%Y-%m-%d')
        dt2 = datetime.strptime(date_range[1], '%Y-%m-%d')

        year = dt1.year
        month = dt1.month

        months = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

        if month < dt2.month:
            month_range = months[months.index(month):months.index(dt2.month)+1]
        else:
            month_range = months[months.index(month):] + months[:months.index(dt2.month)+1]

        while True:

            if year > dt2.year:
                break

            for m in month_range:

                yearmonth_query = '{:d}{:02d}'.format(year, m)

                for sensor in sensors:

                    band_associations = self.associations[sensor]

                    # TODO: get path/row and MGRS from geometry
                    # location = '21/H/UD' # or '225/083'

                    if sensor.lower() == 's2':

                        locations = ['{}/{}/{}'.format(dfrow.Name[:2], dfrow.Name[2], dfrow.Name[3:])
                                     for dfi, dfrow in shp_dict['mgrs'].iterrows()]

                    else:

                        locations = ['{:03d}/{:03d}'.format(int(dfrow.PATH), int(dfrow.ROW))
                                     for dfi, dfrow in shp_dict['wrs'].iterrows()]

                    for location in locations:

                        if sensor.lower() == 's2':

                            query = '{LOCATION}/*{YM}*.SAFE/GRANULE/*'.format(LOCATION=location,
                                                                              YM=yearmonth_query)

                        else:

                            query = '{LOCATION}/*{PATHROW}_{YM}*_T*'.format(LOCATION=location,
                                                                            PATHROW=location.replace('/', ''),
                                                                            YM=yearmonth_query)

                        # Query and list available files on the GCP
                        self.list_gcp(sensor, query)

                        if not self.search_dict:

                            logger.warning('  No results found for {SENSOR} at location {LOC}, year {YEAR:d}, month {MONTH:d}.'.format(SENSOR=sensor,
                                                                                                                                       LOC=location,
                                                                                                                                       YEAR=year,
                                                                                                                                       MONTH=m))

                            continue

                        # Download data
                        if sensor.lower() == 's2':

                            load_bands = sorted(['B{:02d}'.format(band_associations[bd]) if bd != 'rededge' else 'B{:02d}A'.format(band_associations[bd]) for bd in bands])

                            search_wildcards = ['MTD_TL.xml'] + [bd + '.jp2' for bd in load_bands]

                            file_info = self.download_gcp(sensor,
                                                          outdir=outdir,
                                                          search_wildcards=search_wildcards,
                                                          check_file=status.as_posix(),
                                                          verbose=1)

                            # Reorganize the dictionary to combine bands and metadata
                            new_dict_ = dict()
                            for finfo_key, finfo_dict in file_info.items():

                                sub_dict_ = dict()

                                if 'meta' in finfo_dict:

                                    key = finfo_dict['meta'].name

                                    sub_dict_['meta'] = finfo_dict['meta']

                                    for finfo_key_, finfo_dict_ in file_info.items():

                                        if 'meta' not in finfo_dict_:
                                            for bdkey_, bdinfo_ in finfo_dict_.items():
                                                if '_'.join(bdinfo_.name.split('_')[:-1]) in key:
                                                    sub_dict_[bdkey_] = bdinfo_

                                    new_dict_[finfo_key] = sub_dict_

                            file_info = new_dict_

                        else:

                            del_keys = [k for k, v in self.search_dict.items() if 'gap_mask' in k]

                            for dk in del_keys:
                                del self.search_dict[dk]

                            load_bands = sorted(['B{:d}'.format(band_associations[bd]) for bd in bands])

                            search_wildcards = ['ANG.txt', 'MTL.txt', 'BQA.TIF'] + [bd + '.TIF' for bd in load_bands]

                            file_info = self.download_gcp(sensor,
                                                          outdir=outdir,
                                                          search_wildcards=search_wildcards,
                                                          check_file=status.as_posix(),
                                                          verbose=1)

                        logger.info('  Finished downloading files')

                        # Create pixel angle files
                        # TODO: this can be run in parallel
                        for finfo_key, finfo_dict in file_info.items():

                            brdfp = '_'.join(Path(finfo_dict['meta'].name).name.split('_')[:-1])

                            logger.info('  Processing {} ...'.format(brdfp))

                            out_brdf = outdir_brdf.joinpath(brdfp + '.tif').as_posix()
                            out_angles = outdir_brdf.joinpath(brdfp + '_angles.tif').as_posix()

                            if os.path.isfile(out_brdf):
                                logger.warning('  The output BRDF file, {}, already exists.'.format(brdfp))
                                continue

                            with open(status.as_posix(), mode='r') as tx:
                                lines = tx.readlines()

                            if sensor == 's2':
                                outdir_angles = main_path.joinpath('angles_{}'.format(Path(finfo_dict['meta'].name).name.replace('_MTD_TL.xml', '')))
                            else:
                                outdir_angles = main_path.joinpath('angles_{}'.format(Path(finfo_dict['meta'].name).name.replace('_MTL.txt', '')))

                            outdir_angles.mkdir(parents=True, exist_ok=True)

                            ref_file = finfo_dict[load_bands[0]].name

                            if sensor.lower() == 's2':

                                angle_info = sentinel_pixel_angles(finfo_dict['meta'].name,
                                                                   ref_file,
                                                                   outdir_angles.as_posix(),
                                                                   nodata=-32768,
                                                                   overwrite=False,
                                                                   verbose=1)

                                if ' '.join(bands) == 'blue green red nir1 nir2 nir3 nir rededge swir1 swir2':
                                    rad_sensor = 's2'
                                elif ' '.join(bands) == 'blue green red nir swir1 swir2':
                                    rad_sensor = 's2l7'
                                elif ' '.join(bands) == 'blue green red nir':
                                    rad_sensor = 's210'
                                else:
                                    rad_sensor = 's2'

                                bandpass_sensor = angle_info.sensor
                                
                            else:

                                meta = rt.get_landsat_coefficients(finfo_dict['meta'].name)

                                angle_info = landsat_pixel_angles(finfo_dict['angle'].name,
                                                                  ref_file,
                                                                  outdir_angles.as_posix(),
                                                                  meta.sensor,
                                                                  l57_angles_path=l57_angles_path,
                                                                  l8_angles_path=l8_angles_path,
                                                                  verbose=1)

                                if (len(bands) == 1) and (bands[0] == 'pan'):
                                    rad_sensor = sensor + bands[0]
                                else:

                                    if (len(bands) == 6) and (meta.sensor == 'l8'):
                                        rad_sensor = 'l8l7'
                                    elif (len(bands) == 7) and (meta.sensor == 'l8') and ('pan' in bands):
                                        rad_sensor = 'l8l7mspan'
                                    elif (len(bands) == 7) and (meta.sensor == 'l7') and ('pan' in bands):
                                        rad_sensor = 'l7mspan'
                                    else:
                                        rad_sensor = meta.sensor

                                bandpass_sensor = sensor

                            # Get band names from user
                            load_bands_names = [finfo_dict[bd].name for bd in load_bands]

                            with gw.config.update(sensor=rad_sensor,
                                                  ref_bounds=out_bounds,
                                                  ref_crs=crs,
                                                  ref_res=ref_res if ref_res else load_bands_names[-1]):

                                with gw.open(angle_info.sza,
                                             resampling='cubic') as sza, \
                                        gw.open(angle_info.vza,
                                                resampling='cubic') as vza, \
                                        gw.open(angle_info.saa,
                                                resampling='cubic') as saa, \
                                        gw.open(angle_info.vaa,
                                                resampling='cubic') as vaa:

                                    with gw.open(load_bands_names,
                                                 band_names=bands,
                                                 stack_dim='band',
                                                 resampling='cubic') as data:#, \
                                            # gw.open(finfo_dict['qa'].name,
                                            #         band_names=['qa']) as qa:

                                        # Setup the mask
                                        # if sensor.lower() != 's2':
                                        #
                                        #     if sensor.lower() == 'l8':
                                        #         qa_sensor = 'l8-c1'
                                        #     else:
                                        #         qa_sensor = 'l-c1'

                                        # mask = QAMasker(qa,
                                        #                 qa_sensor,
                                        #                 mask_items=['clear',
                                        #                             'fill',
                                        #                             'shadow',
                                        #                             'cloudconf',
                                        #                             'cirrusconf',
                                        #                             'snowiceconf'],
                                        #                 confidence_level='maybe').to_mask()

                                        if sensor.lower() == 's2':

                                            # The S-2 data are in TOAR (0-10000)
                                            toar_scaled = (data * 0.0001).clip(0, 1).astype('float64')
                                            toar_scaled.attrs = data.attrs.copy()

                                            # Convert TOAR to surface reflectance
                                            sr = rt.toar_to_sr(toar_scaled,
                                                               sza, saa, vza, vaa,
                                                               sensor)

                                        else:

                                            # Convert DN to surface reflectance
                                            sr = rt.dn_to_sr(data,
                                                             sza, saa, vza, vaa,
                                                             sensor=rad_sensor,
                                                             meta=meta)

                                        # BRDF normalization
                                        sr_brdf = br.norm_brdf(sr,
                                                               sza, saa, vza, vaa,
                                                               sensor=rad_sensor,
                                                               wavelengths=data.band.values.tolist(),
                                                               out_range=10000.0,
                                                               nodata=65535)

                                        if bandpass_sensor.lower() in ['l5', 'l7', 's2a', 's2b']:

                                            # Linear adjust to Landsat 8
                                            sr_brdf = la.bandpass(sr_brdf,
                                                                  bandpass_sensor.lower(),
                                                                  to='l8',
                                                                  scale_factor=0.0001)

                                        # Mask non-clear pixels
                                        # attrs = sr_brdf.attrs
                                        # sr_brdf = xr.where(mask.sel(band='mask') < 2, sr_brdf, 65535)
                                        # sr_brdf = sr_brdf.transpose('band', 'y', 'x')
                                        # sr_brdf.attrs = attrs

                                        attrs = sr_brdf.attrs.copy()
                                        sr_brdf = sr_brdf.clip(0, 10000).astype('uint16')
                                        sr_brdf.attrs = attrs.copy()

                                        sr_brdf.gw.to_raster(out_brdf, **kwargs)

                                        if write_angle_files:

                                            angle_stack = xr.concat((sza, saa), dim='band')
                                            angle_stack.attrs = sza.attrs.copy()
                                            angle_stack.gw.to_raster(out_angles, **kwargs)

                            angle_infos[finfo_key] = angle_info

                            shutil.rmtree(outdir_angles.as_posix())

                            for k, v in finfo_dict.items():
                                os.remove(v.name)

                            lines.append(finfo_dict['meta'].name + '\n')

                            with open(status.as_posix(), mode='r+') as tx:
                                tx.writelines(lines)

            year += 1

    def list_gcp(self, sensor, query):

        """
        Lists files from Google Cloud Platform

        Args:
            sensor (str): The sensor to query. Choices are ['l5', 'l7', 'l8', 's2'].
            query (str): The query string.

        Examples:
            >>> dl = GeoDownloads()
            >>>
            >>> # Query from a known directory
            >>> dl.list_gcp('landsat', 'LC08/01/042/034/LC08_L1TP_042034_20161104_20170219_01_T1/')
            >>>
            >>> # Query a date for Landsat 5
            >>> dl.list_gcp('l5', '042/034/*2016*')
            >>>
            >>> # Query a date for Landsat 7
            >>> dl.list_gcp('l7', '042/034/*2016*')
            >>>
            >>> # Query a date for Landsat 8
            >>> dl.list_gcp('l8', '042/034/*2016*')
            >>>
            >>> # Query Sentinel-2
            >>> dl.list_gcp('s2', '21/H/UD/*2019*.SAFE/GRANULE/*')

        Returns:
            ``dict``
        """

        gcp_dict = dict(l5='LT05/01',
                        l7='LE07/01',
                        l8='LC08/01',
                        s2='tiles')

        if sensor not in ['l5', 'l7', 'l8', 's2']:
            logger.exception("  The sensor must be 'l5', 'l7', 'l8', or 's2'.")

        if sensor == 's2':
            gcp_str = "gsutil ls -r gs://gcp-public-data-sentinel-2"
        else:
            gcp_str = "gsutil ls -r gs://gcp-public-data-landsat"

        # gsutil_str = '{GSUTIL}/{COLLECTION}/{QUERY}'.format(GSUTIL=gcp_str,
        #                                                     COLLECTION=gcp_dict[sensor],
        #                                                     QUERY=query)

        gsutil_str = gcp_str + "/" + gcp_dict[sensor] + "/" + query

        proc = subprocess.run(gsutil_str.split(' '),
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

        output = proc.stdout

        search_list = [outp for outp in output.decode('utf-8').split('\n') if '$folder$' not in outp]

        if search_list:

            # Check for lenth-1 lists with empty strings
            if search_list[0]:

                if sensor == 's2':
                    self.search_dict = self._prepare_gcp_dict(search_list, 'gs://gcp-public-data-sentinel-2/')
                else:
                    self.search_dict = self._prepare_gcp_dict(search_list, 'gs://gcp-public-data-landsat/')

    @staticmethod
    def _prepare_gcp_dict(search_list, gcp_str):

        """
        Prepares a list of GCP keys into a dictionary

        Args:
            search_list (list)

        Returns:
            ``dict``
        """

        df = pd.DataFrame(data=search_list, columns=['url'])

        df['mask'] = df.url.str.strip().str.endswith('/:')

        mask_idx = np.where(df['mask'].values)[0]
        mask_range = mask_idx.shape[0] - 1 if mask_idx.shape[0] > 1 else 1

        url_dict = dict()

        for mi in range(0, mask_range):

            m1 = mask_idx[mi]

            if mask_range > 1:
                m2 = mask_idx[mi + 1] - 1
            else:
                m2 = len(search_list)

            key = search_list[m1].replace(gcp_str, '').replace('/:', '')
            values = search_list[m1:m2]

            values = [value for value in values if value]

            url_dict[key] = [value for value in values if not value.endswith('/:')]

        return url_dict

    def download_gcp(self,
                     sensor,
                     downloads=None,
                     outdir='.',
                     search_wildcards=None,
                     search_dict=None,
                     check_file=None,
                     verbose=0):

        """
        Downloads a file from Google Cloud platform

        Args:
            sensor (str): The sensor to query. Choices are ['l5', 'l7', 'l8', 's2'].
            downloads (Optional[str or list]): The file or list of keys to download. If not given, keys will be taken
                from ``search_dict`` or ``self.search_dict``.
            outdir (Optional[str]): The output directory.
            search_wildcards (Optional[list]): A list of search wildcards.
            search_dict (Optional[dict]): A keyword search dictionary to override ``self.search_dict``.
            check_file (Optional[str]): A status file to check.
            verbose (Optional[int]): The verbosity level.

        Returns:
            ``namedtuple`` or ``list``
        """

        if not search_dict:

            if not self.search_dict:
                logger.exception('  A keyword search dictionary must be provided, either from `self.list_gcp` or the `search_dict` argument.')
            else:
                search_dict = self.search_dict

        poutdir = Path(outdir)

        if outdir != '.':
            poutdir.mkdir(parents=True, exist_ok=True)

        if not downloads:
            downloads = list(search_dict.keys())

        if not isinstance(downloads, list):
            downloads = [downloads]

        if sensor == 's2':
            gcp_str = 'gsutil cp -r gs://gcp-public-data-sentinel-2'
        else:
            gcp_str = 'gsutil cp -r gs://gcp-public-data-landsat'

        FileInfo = namedtuple('FileInfo', 'name key')

        downloaded = dict()
        null_items = list()

        for search_key in downloads:

            downloaded_sub = dict()

            download_list = self.search_dict[search_key]

            if search_wildcards:

                download_list_ = list()

                for swild in search_wildcards:
                    download_list_ += fnmatch.filter(download_list, '*{}'.format(swild))

                download_list = download_list_

            for fn in download_list:

                rename = False

                fname = Path(fn).name
                down_file = poutdir.joinpath(fname).as_posix()

                if down_file.endswith('_ANG.txt'):
                    fbase = fname.replace('_ANG.txt', '')
                    key = 'angle'
                elif down_file.endswith('_MTL.txt'):
                    fbase = fname.replace('_MTL.txt', '')
                    key = 'meta'
                elif down_file.endswith('MTD_TL.xml'):

                    fbase = Path(fn).parent.name
                    # fsplit = fbase.split('_')
                    # fbase = fsplit[1] + '_' + fsplit[3][:8]
                    down_file = poutdir.joinpath(fbase + '_MTD_TL.xml').as_posix()
                    key = 'meta'
                    rename = True

                elif down_file.endswith('_BQA.TIF'):
                    fbase = fname.replace('_BQA.TIF', '')
                    key = 'qa'
                else:

                    if fname.endswith('.jp2'):

                        fbase = Path(fn).parent.parent.name
                        # fsplit = fbase.split('_')
                        # fbase = fsplit[0] + '_' + fsplit[1][:8]
                        key = Path(fn).name.split('.')[0].split('_')[-1]
                        down_file = poutdir.joinpath(fbase + '_' + key + '.jp2').as_posix()
                        rename = True

                    else:

                        fsplit = fname.split('_')
                        fbase = '_'.join(fsplit[:-1])
                        key = fsplit[-1].split('.')[0]

                continue_download = True

                if fbase in null_items:
                    continue_download = False

                if continue_download and check_file:

                    with open(check_file, mode='r') as tx:
                        lines = tx.readlines()

                    if Path(down_file).parent.joinpath(fbase + '_MTL.txt').as_posix() + '\n' in lines:

                        null_items.append(fbase)
                        continue_download = False

                if continue_download:

                    if not Path(down_file).is_file():

                        if fn.lower().startswith('gs://gcp-public-data'):
                            com = 'gsutil cp -r {} {}'.format(fn, outdir)
                        else:
                            com = 'gsutil cp -r {}/{} {}'.format(gcp_str, fn, outdir)

                        if verbose > 0:
                            logger.info('  Downloading {} ...'.format(fname))

                        subprocess.call(com, shell=True)

                        if rename:
                            os.rename(Path(outdir).joinpath(Path(fn).name).as_posix(), down_file)

                    downloaded_sub[key] = FileInfo(name=down_file, key=key)

            if downloaded_sub:
                downloaded[search_key] = downloaded_sub

        return downloaded

    def download_landsat_range(self, sensors, bands, path_range, row_range, date_range, **kwargs):

        """
        Downloads Landsat data from iterables

        Args:
            sensors (str): A list of sensors to download.
            bands (str): A list of bands to download.
            path_range (iterable): A list of paths.
            row_range (iterable): A list of rows.
            date_range (iterable): A list of ``datetime`` objects or a list of strings as yyyymmdd.
            kwargs (Optional[dict]): Keyword arguments to pass to ``download``.

        Examples:
            >>> from geowombat.util import download_landsat_range
            >>>
            >>> download_landsat_range(['lc08'], ['b4'], [42], [34], ['20170616', '20170620'])
        """

        if (len(date_range) == 2) and not isinstance(date_range[0], datetime):

            start_date = date_range[0]
            end_date = date_range[1]

            sdt = datetime.strptime(start_date, '%Y%m%d')
            edt = datetime.strptime(end_date, '%Y%m%d')

            date_range = pd.date_range(start=sdt, end=edt).to_pydatetime().tolist()

        for sensor in sensors:
            for band in bands:
                for path in path_range:
                    for row in row_range:
                        for dt in date_range:

                            str_date = '{:d}{:02d}{:02d}'.format(dt.year, dt.month, dt.day)

                            # TODO: check if L1TP is used for all sensors
                            # TODO: fixed DATE2
                            filename = '{SENSOR}_L1TP_{PATH:03d}{ROW:03d}_{DATE}_{DATE2}_01_T1_{BAND}.TIF'.format(SENSOR=sensor.upper(),
                                                                                                                  PATH=path,
                                                                                                                  ROW=row,
                                                                                                                  DATE=str_date,
                                                                                                                  DATE2=None,
                                                                                                                  BAND=band)

                            self.download(filename, **kwargs)

    # def download(self, filename, outdir='.', from_google=True, metadata=True, overwrite=False):
    #
    #     """
    #     Downloads an individual file
    #
    #     Args:
    #         filename (str or list): The file to download.
    #         outdir (Optional[str]): The output directory.
    #         from_google (Optional[bool]): Whether to download from Google Cloud storage
    #         metadata (Optional[bool]): Whether to download metadata files.
    #         overwrite (Optional[bool]): Whether to overwrite an existing file.
    #
    #     https://storage.googleapis.com/gcp-public-data-landsat/LC08/01/042/034/LC08_L1TP_042034_20170616_20170629_01_T1/LC08_L1TP_042034_20170616_20170629_01_T1_B4.TIF
    #
    #     Examples:
    #         >>> from geowombat.util import download
    #         >>>
    #         >>> # Download band 4 from Google Cloud storage to the current directory
    #         >>> download('LC08_L1TP_042034_20170616_20170629_01_T1_B4.TIF')
    #
    #     Returns:
    #         None
    #     """
    #
    #     outputs = list()
    #
    #     if not isinstance(filename, list):
    #         filename = [filename]
    #
    #     FileInfo = namedtuple('FileInfo', 'band meta angles')
    #
    #     if outdir != '.':
    #         Path(outdir).mkdir(parents=True, exist_ok=True)
    #
    #     for fn in filename:
    #
    #         if from_google:
    #
    #             file_info = _parse_google_filename(fn,
    #                                                self.landsat_parts,
    #                                                self.sentinel_parts,
    #                                                self.gcp_public)
    #
    #             file_on_disc = Path(outdir).joinpath(fn)
    #             meta_on_disc = Path(outdir).joinpath(Path(file_info.meta).name)
    #             angles_on_disc = Path(outdir).joinpath(Path(file_info.angles).name)
    #
    #             if file_info.url:
    #
    #                 if overwrite:
    #
    #                     if file_on_disc.exists():
    #                         file_on_disc.unlink()
    #
    #                 if file_on_disc.exists():
    #                     logger.warning('  The file already exists.')
    #                 else:
    #
    #                     wget.download(file_info.url_file, out=outdir)
    #
    #                     if metadata:
    #
    #                         wget.download(file_info.meta, out=outdir)
    #                         wget.download(file_info.angles, out=outdir)
    #
    #             outputs.append(FileInfo(band=file_on_disc.as_posix(),
    #                                     meta=meta_on_disc.as_posix(),
    #                                     angles=angles_on_disc.as_posix()))
    #
    #     return outputs



from ..errors import logger

import numpy as np
from osgeo import gdal, gdal_array
import cv2
import dask
import dask.array as da
import xarray as xr
from sklearn.linear_model import LinearRegression, TheilSenRegressor


def calc_slope(elev, proc_dims=None, w=None, **kwargs):

    """
    Calculates slope from elevation

    Args:
        elev (2d array): The elevation data.
        proc_dims (Optional[tuple]): Dimensions to resize to.
        w (Optional[int]): The smoothing window size when ``proc_dims`` is given.
        kwargs (Optional[dict]): Keyword arguments passed to ``gdal.DEMProcessingOptions``.

    Returns:
        ``numpy.ndarray``
    """

    if proc_dims:

        inrows, incols = elev.shape

        elev = cv2.resize(elev.astype('float32'),
                          proc_dims,
                          interpolation=cv2.INTER_LINEAR)

    ds = gdal_array.OpenArray(elev.astype('float64'))

    slope_options = gdal.DEMProcessingOptions(**kwargs)

    out_ds = gdal.DEMProcessing('', ds, 'slope', options=slope_options)

    dst_array = out_ds.GetRasterBand(1).ReadAsArray()

    ds = None
    out_ds = None

    if proc_dims:

        dst_array = cv2.resize(dst_array.astype('float32'),
                               (incols, inrows),
                               interpolation=cv2.INTER_LINEAR)

        if not w:
            w = 15

        return np.float64(cv2.bilateralFilter(np.float32(dst_array), w, 10, 10))

    else:
        return np.float64(dst_array)


def calc_aspect(elev, proc_dims=None, w=None, **kwargs):

    """
    Calculates aspect from elevation

    Args:
        elev (2d array): The elevation data.
        proc_dims (Optional[tuple]): Dimensions to resize to.
        w (Optional[int]): The smoothing window size when ``proc_dims`` is given.
        kwargs (Optional[dict]): Keyword arguments passed to ``gdal.DEMProcessingOptions``.

    Returns:
        ``numpy.ndarray``
    """

    if proc_dims:

        inrows, incols = elev.shape

        elev = cv2.resize(elev.astype('float32'),
                          proc_dims,
                          interpolation=cv2.INTER_LINEAR)

    ds = gdal_array.OpenArray(elev.astype('float64'))

    aspect_options = gdal.DEMProcessingOptions(**kwargs)

    out_ds = gdal.DEMProcessing('', ds, 'aspect', options=aspect_options)

    dst_array = out_ds.GetRasterBand(1).ReadAsArray()

    ds = None
    out_ds = None

    if proc_dims:

        dst_array = cv2.resize(dst_array.astype('float32'),
                               (incols, inrows),
                               interpolation=cv2.INTER_LINEAR)

        if not w:
            w = 15

        return np.float64(cv2.bilateralFilter(np.float32(dst_array), w, 10, 10))

    else:
        return np.float64(dst_array)


calc_slope_delayed = dask.delayed(calc_slope)
calc_aspect_delayed = dask.delayed(calc_aspect)


class Topo(object):

    """
    A class for topographic normalization
    """

    def _regress_a(self, X, y, robust, n_jobs):

        """
        Calculates the slope and intercept
        """

        if robust:
            model = TheilSenRegressor(n_jobs=n_jobs)
        else:
            model = LinearRegression(n_jobs=n_jobs)

        model.fit(X, y)

        slope_m = model.coef_[0]
        intercept_b = model.intercept_

        return slope_m, intercept_b

    def _method_empirical_rotation(self, sr, il, cos_z, nodata_samps, n_jobs=1, robust=False):

        r"""
        Normalizes terrain using the Empirical Rotation method

        Args:
            sr (Dask Array): The surface reflectance data.
            il (Dask Array): The solar illumination.
            cos_z (Dask Array): The cosine of the solar zenith angle.
            nodata_samps (Dask Array): Samples where 1='no data' and 0='valid data'.
            n_jobs (Optional[int]): The number of parallel workers for ``LinearRegression.fit`` or
                ``TheilSenRegressor.fit``.
            robust (Optional[bool]): Whether to fit a robust regression.

        References:

            See :cite:`tan_etal_2010` for the Empirical Rotation method.

        Returns:
            ``dask.array``
        """

        nodata = nodata_samps.compute().flatten()
        idx = np.where(nodata == 0)[0]

        X = il.compute().flatten()[idx][:, np.newaxis]
        y = sr.compute().flatten()[idx]

        slope_m, intercept_b = self._regress_a(X, y, robust, n_jobs)

        # https://reader.elsevier.com/reader/sd/pii/S0034425713001673?token=6C93FB2E69ABF5729CE9ECBBDFD9C2D985613156753C822A3D160102D46135E01457EE33500DB4648C6AF636F39D8B62
        # Improved forest change detection with terrain illumination corrected Landsat images
        sr_a = sr - slope_m * (il - cos_z)
        # sr_a = sr - (slope_m * il + intercept_b)

        return da.where(nodata_samps == 1, sr, sr_a).clip(0, 1)

    def _method_cos(self, sr, il, cos_z, nodata_samps):

        r"""
        Normalizes terrain using the Cosine method

        Args:
            sr (Dask Array): The surface reflectance data.
            il (Dask Array): The solar illumination.
            cos_z (Dask Array): The cosine of the solar zenith angle.
            nodata_samps (Dask Array): Samples where 1='no data' and 0='valid data'.

        References:

            See :cite:`teillet_etal_1982` for the C-correction method.

        Returns:
            ``dask.array``
        """

        sr_a = sr * (cos_z / il)

        return da.where(nodata_samps == 1, sr, sr_a).clip(0, 1)

    def _method_c(self, sr, il, cos_z, nodata_samps, n_jobs=1, robust=False):

        r"""
        Normalizes terrain using the C-correction method

        Args:
            sr (Dask Array): The surface reflectance data.
            il (Dask Array): The solar illumination.
            cos_z (Dask Array): The cosine of the solar zenith angle.
            nodata_samps (Dask Array): Samples where 1='no data' and 0='valid data'.
            n_jobs (Optional[int]): The number of parallel workers for ``LinearRegression.fit`` or
                ``TheilSenRegressor.fit``.
            robust (Optional[bool]): Whether to fit a robust regression.

        References:

            See :cite:`teillet_etal_1982` for the C-correction method.

        Returns:
            ``dask.array``
        """

        nodata = nodata_samps.compute().flatten()
        idx = np.where(nodata == 0)[0]

        X = il.compute().flatten()[idx][:, np.newaxis]
        y = sr.compute().flatten()[idx]

        slope_m, intercept_b = self._regress_a(X, y, robust, n_jobs)

        c = intercept_b / slope_m

        # Get the A-factor
        a_factor = (cos_z + c) / (il + c)

        a_factor = da.where(da.isnan(a_factor), 1, a_factor)

        sr_a = sr * a_factor

        return da.where((sr_a > 1) | (nodata_samps == 1), sr, sr_a).clip(0, 1)

    def norm_topo(self,
                  data,
                  elev,
                  solar_za,
                  solar_az,
                  slope=None,
                  aspect=None,
                  method='empirical-rotation',
                  slope_thresh=2,
                  nodata=0,
                  elev_nodata=-32768,
                  scale_factor=1,
                  angle_scale=0.01,
                  n_jobs=1,
                  robust=False,
                  slope_kwargs=None,
                  aspect_kwargs=None):

        """
        Applies topographic normalization

        Args:
            data (2d or 3d DataArray): The data to normalize, in the range 0-1.
            elev (2d DataArray): The elevation data.
            solar_za (2d DataArray): The solar zenith angles (degrees).
            solar_az (2d DataArray): The solar azimuth angles (degrees).
            slope (2d DataArray): The slope data. If not given, slope is calculated from ``elev``.
            aspect (2d DataArray): The aspect data. If not given, aspect is calculated from ``elev``.
            method (Optional[str]): The method to apply. Choices are ['c', 'empirical-rotation'].
            slope_thresh (Optional[float or int]): The slope threshold. Any samples with
                values < ``slope_thresh`` are not adjusted.
            nodata (Optional[int or float]): The 'no data' value for ``data``.
            elev_nodata (Optional[float or int]): The 'no data' value for ``elev``.
            scale_factor (Optional[float]): A scale factor to apply to the input data.
            angle_scale (Optional[float]): The angle scale factor.
            n_jobs (Optional[int]): The number of parallel workers for ``LinearRegression.fit``.
            robust (Optional[bool]): Whether to fit a robust regression.
            slope_kwargs (Optional[dict]): Keyword arguments passed to ``gdal.DEMProcessingOptions``
                to calculate the slope.
            aspect_kwargs (Optional[dict]): Keyword arguments passed to ``gdal.DEMProcessingOptions``
                to calculate the aspect.

        References:

            See :cite:`teillet_etal_1982` for the C-correction method.
            See :cite:`tan_etal_2010` for the Empirical Rotation method.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>> from geowombat.radiometry import Topo
            >>>
            >>> topo = Topo()
            >>>
            >>> # Example where pixel angles are stored in separate GeoTiff files
            >>> with gw.config.update(sensor='l7', scale_factor=0.0001, nodata=0):
            >>>
            >>>     with gw.open('landsat.tif') as src,
            >>>         gw.open('srtm') as elev,
            >>>             gw.open('solarz.tif') as solarz,
            >>>                 gw.open('solara.tif') as solara:
            >>>
            >>>         src_norm = topo.norm_topo(src, elev, solarz, solara, n_jobs=-1)
        """

        method = method.strip().lower()

        if method not in ['c', 'empirical-rotation']:

            logger.exception("  Currently, the only supported methods are 'c' and 'empirical-rotation'.")
            raise NameError

        attrs = data.attrs.copy()

        if not nodata:
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        # Scale the reflectance data
        if scale_factor != 1:
            data = data * scale_factor

        if not slope_kwargs:

            slope_kwargs = dict(format='MEM',
                                computeEdges=True,
                                alg='ZevenbergenThorne',
                                slopeFormat='degree')

        if not aspect_kwargs:

            aspect_kwargs = dict(format='MEM',
                                 computeEdges=True,
                                 alg='ZevenbergenThorne',
                                 trigonometric=False,
                                 zeroForFlat=True)

        slope_kwargs['format'] = 'MEM'
        slope_kwargs['slopeFormat'] = 'degree'
        aspect_kwargs['format'] = 'MEM'

        # Force to SRTM resolution
        proc_dims = (int((data.gw.ncols*data.gw.cellx) / 30.0),
                     int((data.gw.nrows*data.gw.celly) / 30.0))

        w = int((5 * 30.0) / data.gw.celly)

        if w % 2 == 0:
            w += 1

        if isinstance(slope, xr.DataArray):
            slope_deg_fd = slope.squeeze().data
        else:

            slope_deg = calc_slope_delayed(elev.squeeze().data, proc_dims=proc_dims, w=w, **slope_kwargs)
            slope_deg_fd = da.from_delayed(slope_deg, (data.gw.nrows, data.gw.ncols), dtype='float64')

        if isinstance(aspect, xr.DataArray):
            aspect_deg_fd = aspect.squeeze().data
        else:

            aspect_deg = calc_aspect_delayed(elev.squeeze().data, proc_dims=proc_dims, w=w, **aspect_kwargs)
            aspect_deg_fd = da.from_delayed(aspect_deg, (data.gw.nrows, data.gw.ncols), dtype='float64')

        nodata_samps = da.where((elev.data == elev_nodata) |
                                (data.max(dim='band').data == nodata) |
                                (slope_deg_fd < slope_thresh), 1, 0)

        slope_rad = da.deg2rad(slope_deg_fd)
        aspect_rad = da.deg2rad(aspect_deg_fd)

        # Convert degrees to radians
        solar_za = da.deg2rad(solar_za.squeeze().data * angle_scale)
        solar_az = da.deg2rad(solar_az.squeeze().data * angle_scale)

        cos_z = da.cos(solar_za)

        # Calculate the illumination angle
        il = da.cos(slope_rad) * cos_z + da.sin(slope_rad) * da.sin(solar_za) * da.cos(solar_az - aspect_rad)

        sr_adj = list()
        for band in data.band.values.tolist():

            if method == 'c':

                sr_adj.append(self._method_c(data.sel(band=band).data,
                                             il,
                                             cos_z,
                                             nodata_samps,
                                             n_jobs=n_jobs,
                                             robust=robust))

            else:

                sr_adj.append(self._method_empirical_rotation(data.sel(band=band).data,
                                                              il,
                                                              cos_z,
                                                              nodata_samps,
                                                              n_jobs=n_jobs,
                                                              robust=robust))

        adj_data = xr.DataArray(data=da.concatenate(sr_adj).reshape((data.gw.nbands,
                                                                     data.gw.nrows,
                                                                     data.gw.ncols)),
                                coords={'band': data.band.values.tolist(),
                                        'y': data.y.values,
                                        'x': data.x.values},
                                dims=('band', 'y', 'x'),
                                attrs=data.attrs)

        attrs['calibration'] = 'Topographic-adjusted'
        attrs['nodata'] = nodata
        attrs['drange'] = (0, 1)

        adj_data.attrs = attrs

        return adj_data



from .util import n_rows_cols

from rasterio.windows import Window


def get_window_offsets(n_rows, n_cols, row_chunks, col_chunks, return_as='list'):

    """
    Gets window offset indices from image dimensions and chunk sizes

    Args:
        n_rows (int): The number of rows to iterate over.
        n_cols (int): The number of columns to iterate over.
        row_chunks (int): The row chunk size.
        col_chunks (int): The column chunk size.
        return_as (Optional[str]): How to return the window information. Choices are ['dict', 'list'].

    Returns:
        Window information (list or dict)
    """

    if return_as == 'list':
        window_info = list()
    else:
        window_info = dict()

    i = 0

    for row_off in range(0, n_rows, row_chunks):

        height = n_rows_cols(row_off, row_chunks, n_rows)

        j = 0

        for col_off in range(0, n_cols, col_chunks):

            width = n_rows_cols(col_off, col_chunks, n_cols)

            if return_as == 'list':

                window_info.append(Window(col_off=col_off,
                                          row_off=row_off,
                                          width=width,
                                          height=height))

            else:

                window_info['{:d}{:d}'.format(i, j)] = Window(col_off=col_off,
                                                              row_off=row_off,
                                                              width=width,
                                                              height=height)

            j += 1

        i += 1

    return window_info



import os

from ..core.windows import get_window_offsets
from ..core.util import parse_filename_dates
from ..errors import logger
from ..config import config
from .rasterio_ import get_ref_image_meta, warp, warp_images, get_file_bounds
from .rasterio_ import transform_crs as rio_transform_crs

import numpy as np
import dask.array as da
import xarray as xr
from xarray.ufuncs import maximum as xr_maximum
from xarray.ufuncs import minimum as xr_mininum
from deprecated import deprecated


def _update_kwarg(ref_obj, ref_kwargs, key):

    """
    Updates keyword arguments for global config parameters

    Args:
        ref_obj (str or object)
        ref_kwargs (dict)
        key (str)

    Returns:
        ``dict``
    """

    if isinstance(ref_obj, str) and os.path.isfile(ref_obj):

        # Get the metadata from the reference image
        ref_meta = get_ref_image_meta(ref_obj)
        ref_kwargs[key] = getattr(ref_meta, key)

    else:

        if ref_obj:
            ref_kwargs[key] = ref_obj

    return ref_kwargs


def _get_raster_coords(filename):

    with xr.open_rasterio(filename) as src:

        x = src.x.values - src.res[0] / 2.0
        y = src.y.values + src.res[1] / 2.0

    return x, y


def _check_config_globals(filenames, bounds_by, ref_kwargs):

    """
    Checks global configuration parameters

    Args:
        filenames (str or str list)
        bounds_by (str)
        ref_kwargs (dict)
    """

    # Check if there is a reference image
    if config['ref_image']:

        if isinstance(config['ref_image'], str) and os.path.isfile(config['ref_image']):

            # Get the metadata from the reference image
            ref_meta = get_ref_image_meta(config['ref_image'])

            ref_kwargs['bounds'] = ref_meta.bounds
            ref_kwargs['crs'] = ref_meta.crs
            ref_kwargs['res'] = ref_meta.res

        else:
            logger.warning('  The reference image does not exist')

    else:

        if config['ref_bounds']:
            ref_kwargs = _update_kwarg(config['ref_bounds'], ref_kwargs, 'bounds')
        else:

            if isinstance(filenames, str):

                # Use the bounds of the image
                ref_kwargs['bounds'] = get_file_bounds([filenames],
                                                       bounds_by='reference',
                                                       crs=ref_kwargs['crs'],
                                                       res=ref_kwargs['res'],
                                                       return_bounds=True)

            else:

                # Replace the bounds keyword, if needed
                if bounds_by.lower() == 'intersection':

                    # Get the intersecting bounds of all images
                    ref_kwargs['bounds'] = get_file_bounds(filenames,
                                                           bounds_by='intersection',
                                                           crs=ref_kwargs['crs'],
                                                           res=ref_kwargs['res'],
                                                           return_bounds=True)

                elif bounds_by.lower() == 'union':

                    # Get the union bounds of all images
                    ref_kwargs['bounds'] = get_file_bounds(filenames,
                                                           bounds_by='union',
                                                           crs=ref_kwargs['crs'],
                                                           res=ref_kwargs['res'],
                                                           return_bounds=True)

                elif bounds_by.lower() == 'reference':

                    # Use the bounds of the first image
                    ref_kwargs['bounds'] = get_file_bounds(filenames,
                                                           bounds_by='reference',
                                                           crs=ref_kwargs['crs'],
                                                           res=ref_kwargs['res'],
                                                           return_bounds=True)

                else:
                    logger.exception("  Choose from 'intersection', 'union', or 'reference'.")

                config['ref_bounds'] = ref_kwargs['bounds']

        if config['ref_crs']:
            ref_kwargs = _update_kwarg(config['ref_crs'], ref_kwargs, 'crs')

        if config['ref_res']:
            ref_kwargs = _update_kwarg(config['ref_res'], ref_kwargs, 'res')

        if config['ref_tar']:

            if isinstance(config['ref_tar'], str):

                if os.path.isfile(config['ref_tar']):
                    ref_kwargs = _update_kwarg(_get_raster_coords(config['ref_tar']), ref_kwargs, 'tac')
                else:
                    logger.warning('  The target aligned raster does not exist.')

            else:
                logger.warning('  The target aligned raster must be an image.')

    return ref_kwargs


def warp_open(filename,
              band_names=None,
              resampling='nearest',
              dtype=None,
              return_windows=False,
              warp_mem_limit=512,
              num_threads=1,
              tap=False,
              **kwargs):

    """
    Warps and opens a file

    Args:
        filename (str): The file to open.
        band_names (Optional[int, str, or list]): The band names.
        resampling (Optional[str]): The resampling method.
        dtype (Optional[str]): A data type to force the output to. If not given, the data type is extracted
            from the file.
        return_windows (Optional[bool]): Whether to return block windows.
        warp_mem_limit (Optional[int]): The memory limit (in MB) for the ``rasterio.vrt.WarpedVRT`` function.
        num_threads (Optional[int]): The number of warp worker threads.
        tap (Optional[bool]): Whether to target align pixels.
        kwargs (Optional[dict]): Keyword arguments passed to ``xarray.open_rasterio``.

    Returns:
        ``xarray.DataArray``
    """

    ref_kwargs = {'bounds': None,
                  'crs': None,
                  'res': None,
                  'warp_mem_limit': warp_mem_limit,
                  'num_threads': num_threads,
                  'tap': tap,
                  'tac': None}

    ref_kwargs = _check_config_globals(filename, 'reference', ref_kwargs)

    with xr.open_rasterio(warp(filename,
                               resampling=resampling,
                               **ref_kwargs),
                          **kwargs) as src:

        if band_names:
            src.coords['band'] = band_names
        else:

            if src.gw.sensor:

                if src.gw.sensor not in src.gw.avail_sensors:

                    logger.warning('  The {} sensor is not currently supported.\nChoose from [{}].'.format(src.gw.sensor,
                                                                                                           ', '.join(src.gw.avail_sensors)))

                else:

                    new_band_names = list(src.gw.wavelengths[src.gw.sensor]._fields)

                    # Avoid nested opens within a `config` context
                    if len(new_band_names) != len(src.band.values.tolist()):

                        logger.warning('  The new bands, {}, do not match the sensor bands, {}.'.format(new_band_names,
                                                                                                        src.band.values.tolist()))

                    else:

                        src.coords['band'] = new_band_names
                        src.attrs['sensor'] = src.gw.sensor_names[src.gw.sensor]

        if return_windows:

            if isinstance(kwargs['chunks'], tuple):
                chunksize = kwargs['chunks'][-1]
            else:
                chunksize = kwargs['chunks']

            src.attrs['block_windows'] = get_window_offsets(src.shape[-2],
                                                            src.shape[-1],
                                                            chunksize,
                                                            chunksize,
                                                            return_as='list')

        src.attrs['filename'] = filename
        src.attrs['resampling'] = resampling

        if dtype:

            attrs = src.attrs.copy()
            return src.astype(dtype).assign_attrs(**attrs)

        else:
            return src


def mosaic(filenames,
           overlap='max',
           bounds_by='reference',
           resampling='nearest',
           band_names=None,
           dtype=None,
           warp_mem_limit=512,
           num_threads=1,
           **kwargs):

    """
    Mosaics a list of images

    Args:
        filenames (list): A list of file names to mosaic.
        overlap (Optional[str]): The keyword that determines how to handle overlapping data.
            Choices are ['min', 'max', 'mean'].
        bounds_by (Optional[str]): How to concatenate the output extent. Choices are ['intersection', 'union', 'reference'].

            * reference: Use the bounds of the reference image
            * intersection: Use the intersection (i.e., minimum extent) of all the image bounds
            * union: Use the union (i.e., maximum extent) of all the image bounds

        resampling (Optional[str]): The resampling method.
        band_names (Optional[1d array-like]): A list of names to give the band dimension.
        dtype (Optional[str]): A data type to force the output to. If not given, the data type is extracted
            from the file.
        warp_mem_limit (Optional[int]): The memory limit (in MB) for the ``rasterio.vrt.WarpedVRT`` function.
        num_threads (Optional[int]): The number of warp worker threads.
        kwargs (Optional[dict]): Keyword arguments passed to ``xarray.open_rasterio``.

    Returns:
        ``xarray.DataArray``
    """

    if overlap not in ['min', 'max', 'mean']:
        logger.exception("  The overlap argument must be one of ['min', 'max', 'mean'].")

    ref_kwargs = {'bounds': None,
                  'crs': None,
                  'res': None,
                  'warp_mem_limit': warp_mem_limit,
                  'num_threads': num_threads,
                  'tac': None}

    ref_kwargs = _check_config_globals(filenames, bounds_by, ref_kwargs)

    # Warp all images to the same grid.
    warped_objects = warp_images(filenames,
                                 bounds_by=bounds_by,
                                 resampling=resampling,
                                 **ref_kwargs)

    # Combine the data
    with xr.open_rasterio(warped_objects[0], **kwargs) as ds:

        attrs = ds.attrs.copy()

        for fn in warped_objects[1:]:

            with xr.open_rasterio(fn, **kwargs) as dsb:

                if overlap == 'min':
                    ds = xr_mininum(ds, dsb)
                elif overlap == 'max':
                    ds = xr_maximum(ds, dsb)
                elif overlap == 'mean':
                    ds = (ds + dsb) / 2.0

                # ds = ds.combine_first(dsb)

        ds = ds.assign_attrs(**attrs)

        if band_names:
            ds.coords['band'] = band_names
        else:

            if ds.gw.sensor:

                if ds.gw.sensor not in ds.gw.avail_sensors:

                    logger.warning('  The {} sensor is not currently supported.\nChoose from [{}].'.format(ds.gw.sensor,
                                                                                                           ', '.join(ds.gw.avail_sensors)))

                else:

                    new_band_names = list(ds.gw.wavelengths[ds.gw.sensor]._fields)

                    if len(new_band_names) != len(ds.band.values.tolist()):
                        logger.warning('  The band list length does not match the sensor bands.')
                    else:

                        ds.coords['band'] = new_band_names
                        ds.attrs['sensor'] = ds.gw.sensor_names[ds.gw.sensor]

        ds.attrs['resampling'] = resampling

        if dtype:

            attrs = ds.attrs.copy()
            return ds.astype(dtype).assign_attrs(**attrs)

        else:
            return ds


def concat(filenames,
           stack_dim='time',
           bounds_by='reference',
           resampling='nearest',
           time_names=None,
           band_names=None,
           dtype=None,
           overlap='max',
           warp_mem_limit=512,
           num_threads=1,
           tap=False,
           **kwargs):

    """
    Concatenates a list of images

    Args:
        filenames (list): A list of file names to concatenate.
        stack_dim (Optional[str]): The stack dimension. Choices are ['time', 'band'].
        bounds_by (Optional[str]): How to concatenate the output extent. Choices are ['intersection', 'union', 'reference'].

            * reference: Use the bounds of the reference image
            * intersection: Use the intersection (i.e., minimum extent) of all the image bounds
            * union: Use the union (i.e., maximum extent) of all the image bounds

        resampling (Optional[str]): The resampling method.
        time_names (Optional[1d array-like]): A list of names to give the time dimension.
        band_names (Optional[1d array-like]): A list of names to give the band dimension.
        dtype (Optional[str]): A data type to force the output to. If not given, the data type is extracted
            from the file.
        overlap (Optional[str]): The keyword that determines how to handle overlapping data.
            Choices are ['min', 'max', 'mean']. Only used when mosaicking arrays from the same timeframe.
        warp_mem_limit (Optional[int]): The memory limit (in MB) for the ``rasterio.vrt.WarpedVRT`` function.
        num_threads (Optional[int]): The number of warp worker threads.
        tap (Optional[bool]): Whether to target align pixels.
        kwargs (Optional[dict]): Keyword arguments passed to ``xarray.open_rasterio``.

    Returns:
        ``xarray.DataArray``
    """

    if stack_dim.lower() not in ['band', 'time']:
        logger.exception("  The stack dimension should be 'band' or 'time'")

    ref_kwargs = {'bounds': None,
                  'crs': None,
                  'res': None,
                  'warp_mem_limit': warp_mem_limit,
                  'num_threads': num_threads,
                  'tap': tap,
                  'tac': None}

    ref_kwargs = _check_config_globals(filenames, bounds_by, ref_kwargs)

    # Keep a copy of the transformed attributes.
    with xr.open_rasterio(warp(filenames[0],
                               resampling=resampling,
                               **ref_kwargs), **kwargs) as ds_:

        attrs = ds_.attrs.copy()

    if time_names:

        concat_list = list()
        new_time_names = list()

        # Check the time names for duplicates
        for tidx in range(0, len(time_names)):

            if list(time_names).count(time_names[tidx]) > 1:

                if time_names[tidx] not in new_time_names:

                    # Get the file names to mosaic
                    filenames_mosaic = [filenames[i] for i in range(0, len(time_names))
                                        if time_names[i] == time_names[tidx]]

                    # Mosaic the images into a single-date array
                    concat_list.append(mosaic(filenames_mosaic,
                                              overlap=overlap,
                                              bounds_by=bounds_by,
                                              resampling=resampling,
                                              band_names=band_names,
                                              warp_mem_limit=warp_mem_limit,
                                              num_threads=num_threads,
                                              **kwargs))

                    new_time_names.append(time_names[tidx])

            else:

                new_time_names.append(time_names[tidx])

                # Warp the date
                concat_list.append(warp_open(filenames[tidx],
                                             resampling=resampling,
                                             band_names=band_names,
                                             warp_mem_limit=warp_mem_limit,
                                             num_threads=num_threads,
                                             **kwargs))

        # Warp all images and concatenate along the 'time' axis into a DataArray
        output = xr.concat(concat_list, dim=stack_dim.lower())

        # Assign the new time band names
        ds = output.assign_coords(time=new_time_names)

    else:

        # Warp all images and concatenate along
        #   the 'time' axis into a DataArray.
        ds = xr.concat([xr.open_rasterio(warp(fn,
                                              resampling=resampling,
                                              **ref_kwargs), **kwargs)
                        for fn in filenames], dim=stack_dim.lower())

    ds = ds.assign_attrs(**attrs)

    if not time_names and (stack_dim == 'time'):
        ds.coords['time'] = parse_filename_dates(filenames)

    if band_names:
        ds.coords['band'] = band_names
    else:

        if ds.gw.sensor:

            if ds.gw.sensor not in ds.gw.avail_sensors:

                logger.warning('  The {} sensor is not currently supported.\nChoose from [{}].'.format(ds.gw.sensor,
                                                                                                       ', '.join(ds.gw.avail_sensors)))

            else:

                new_band_names = list(ds.gw.wavelengths[ds.gw.sensor]._fields)

                if len(new_band_names) != len(ds.band.values.tolist()):

                    logger.warning('  The new bands, {}, do not match the sensor bands, {}.'.format(new_band_names,
                                                                                                    ds.band.values.tolist()))

                else:

                    ds.coords['band'] = new_band_names
                    ds.attrs['sensor'] = ds.gw.sensor_names[ds.gw.sensor]

    if dtype:
        
        attrs = ds.attrs.copy()
        return ds.astype(dtype).assign_attrs(**attrs)

    else:
        return ds


def transform_crs(data_src,
                  dst_crs=None,
                  dst_res=None,
                  dst_width=None,
                  dst_height=None,
                  dst_bounds=None,
                  resampling='nearest',
                  warp_mem_limit=512,
                  num_threads=1):

    """
    Transforms a DataArray to a new coordinate reference system

    Args:
        data_src (DataArray): The data to transform.
        dst_crs (Optional[CRS | int | dict | str]): The destination CRS.
        dst_res (Optional[tuple]): The destination resolution.
        dst_width (Optional[int]): The destination width. Cannot be used with ``dst_res``.
        dst_height (Optional[int]): The destination height. Cannot be used with ``dst_res``.
        dst_bounds (Optional[BoundingBox | tuple]): The destination bounds, as a ``rasterio.coords.BoundingBox``
            or as a tuple of (left, bottom, right, top).
        resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
            Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        warp_mem_limit (Optional[int]): The warp memory limit.
        num_threads (Optional[int]): The number of parallel threads.

    Returns:
        ``xarray.DataArray``
    """

    data_dst, dst_transform, dst_crs = rio_transform_crs(data_src,
                                                         dst_crs=dst_crs,
                                                         dst_res=dst_res,
                                                         dst_width=dst_width,
                                                         dst_height=dst_height,
                                                         dst_bounds=dst_bounds,
                                                         resampling=resampling,
                                                         warp_mem_limit=warp_mem_limit,
                                                         num_threads=num_threads)

    nrows, ncols = data_dst.shape[-2], data_dst.shape[-1]

    left = dst_transform[2]
    cellx = abs(dst_transform[0])
    x = np.arange(left + cellx / 2.0, left + cellx / 2.0 + (cellx * ncols), cellx)

    top = dst_transform[5]
    celly = abs(dst_transform[4])
    y = np.arange(top - celly / 2.0, top - celly / 2.0 - (celly * nrows), -celly)

    if not dst_res:
        dst_res = (abs(x[1] - x[0]), abs(y[0] - y[1]))

    data_dst = xr.DataArray(data=da.from_array(data_dst,
                                               chunks=data_src.data.chunksize),
                            coords={'band': data_src.band.values.tolist(),
                                    'y': y,
                                    'x': x},
                            dims=('band', 'y', 'x'),
                            attrs=data_src.attrs)

    data_dst.attrs['transform'] = tuple(dst_transform)[:6]
    data_dst.attrs['crs'] = dst_crs
    data_dst.attrs['res'] = dst_res
    data_dst.attrs['resampling'] = resampling

    if 'sensor' in data_src.attrs:
        data_dst.attrs['sensor'] = data_src.attrs['sensor']

    if 'filename' in data_src.attrs:
        data_dst.attrs['filename'] = data_src.attrs['filename']

    return data_dst


@deprecated('Deprecated since 1.2.0. Use geowombat.transform_crs() instead.')
def to_crs(data_src,
           dst_crs=None,
           dst_res=None,
           dst_width=None,
           dst_height=None,
           dst_bounds=None,
           resampling='nearest',
           warp_mem_limit=512,
           num_threads=1):

    """
    .. deprecated:: 1.2.0
        Use :func:`geowombat.transform_crs()` instead.

    Transforms a DataArray to a new coordinate reference system

    Args:
        data_src (DataArray): The data to transform.
        dst_crs (Optional[CRS | int | dict | str]): The destination CRS.
        dst_res (Optional[tuple]): The destination resolution.
        dst_width (Optional[int]): The destination width. Cannot be used with ``dst_res``.
        dst_height (Optional[int]): The destination height. Cannot be used with ``dst_res``.
        dst_bounds (Optional[BoundingBox | tuple]): The destination bounds, as a ``rasterio.coords.BoundingBox``
            or as a tuple of (left, bottom, right, top).
        resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
            Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        warp_mem_limit (Optional[int]): The warp memory limit.
        num_threads (Optional[int]): The number of parallel threads.

    Returns:
        ``xarray.DataArray``
    """

    data_dst, dst_transform, dst_crs = rio_transform_crs(data_src,
                                                         dst_crs=dst_crs,
                                                         dst_res=dst_res,
                                                         dst_width=dst_width,
                                                         dst_height=dst_height,
                                                         dst_bounds=dst_bounds,
                                                         resampling=resampling,
                                                         warp_mem_limit=warp_mem_limit,
                                                         num_threads=num_threads)

    nrows, ncols = data_dst.shape[-2], data_dst.shape[-1]

    left = dst_transform[2]
    cellx = abs(dst_transform[0])
    x = np.arange(left + cellx / 2.0, left + cellx / 2.0 + (cellx * ncols), cellx)

    top = dst_transform[5]
    celly = abs(dst_transform[4])
    y = np.arange(top - celly / 2.0, top - celly / 2.0 - (celly * nrows), -celly)

    if not dst_res:
        dst_res = (abs(x[1] - x[0]), abs(y[0] - y[1]))

    data_dst = xr.DataArray(data=da.from_array(data_dst,
                                               chunks=data_src.data.chunksize),
                            coords={'band': data_src.band.values.tolist(),
                                    'y': y,
                                    'x': x},
                            dims=('band', 'y', 'x'),
                            attrs=data_src.attrs)

    data_dst.attrs['transform'] = tuple(dst_transform)[:6]
    data_dst.attrs['crs'] = dst_crs
    data_dst.attrs['res'] = dst_res
    data_dst.attrs['resampling'] = resampling

    if 'sensor' in data_src.attrs:
        data_dst.attrs['sensor'] = data_src.attrs['sensor']

    if 'filename' in data_src.attrs:
        data_dst.attrs['filename'] = data_src.attrs['filename']

    return data_dst



import os
import fnmatch
from collections import namedtuple, OrderedDict
from datetime import datetime

from ..errors import logger
from ..moving import moving_window

import numpy as np
import geopandas as gpd
import xarray as xr
import dask.array as da

from rasterio import features
from rasterio.crs import CRS
from rasterio.warp import reproject, transform_bounds
from rasterio.transform import from_bounds

import shapely
from affine import Affine
from dateparser.search import search_dates

shapely.speedups.enable()


def lazy_wombat(func):
    func.wombat_func_ = True
    return func


def parse_filename_dates(filenames):

    """
    Parses dates from file names

    Args:
        filenames (list): A list of files to parse.

    Returns:
        ``list``
    """

    date_filenames = list()

    for fn in filenames:

        d_name, f_name = os.path.split(fn)
        f_base, f_ext = os.path.splitext(f_name)

        try:

            s, dt = list(zip(*search_dates(' '.join(' '.join(f_base.split('_')).split('-')),
                                           settings={'DATE_ORDER': 'YMD',
                                                     'STRICT_PARSING': False,
                                                     'PREFER_LANGUAGE_DATE_ORDER': False})))

        except:
            return list(range(1, len(filenames) + 1))

        if not dt:
            return list(range(1, len(filenames) + 1))

        date_filenames.append(dt[0])

    return date_filenames


def parse_wildcard(string):

    """
    Parses a search wildcard from a string

    Args:
        string (str): The string to parse.

    Returns:
        ``list``
    """

    if os.path.dirname(string):
        d_name, wildcard = os.path.split(string)
    else:

        d_name = '.'
        wildcard = string

    matches = sorted(fnmatch.filter(os.listdir(d_name), wildcard))

    if matches:
        matches = [os.path.join(d_name, fn) for fn in matches]

    if not matches:
        logger.exception('  There were no images found with the string search.')

    return matches


def sort_images_by_date(image_path,
                        image_wildcard,
                        date_pos,
                        date_start,
                        date_end,
                        split_by='_',
                        date_format='%Y%m%d'):

    """
    Sorts images by date

    Args:
        image_path (Path): The image directory.
        image_wildcard (str): The image search wildcard.
        date_pos (int): The date starting position in the file name.
        date_start (int): The date starting position in the split.
        date_end (int): The date ending position in the split.
        split_by (Optional[str]): How to split the file name.
        date_format (Optional[str]): The date format for ``datetime.datetime.strptime``.

    Returns:
        ``collections.OrderedDict``

    Example:
        >>> from geowombat.core import sort_images
        >>>
        >>> # image example: LC08_L1TP_176038_20190108_20190130_01_T1.tif
        >>> image_path = '/path/to/images'
        >>>
        >>> image_dict = sort_images(image_path, '*.tif', 3, 0, 8)
        >>> image_names = list(image_dict.keys())
        >>> time_names = list(image_dict.values())
    """

    fl = fnmatch.filter(os.listdir(image_path.as_posix()), image_wildcard)
    fl = [image_path.joinpath(fn) for fn in fl]
    dates = [datetime.strptime(fn.name.split(split_by)[date_pos][date_start:date_end], date_format) for fn in fl]

    return OrderedDict(sorted(dict(zip([fn.as_posix() for fn in fl], dates)).items(), key=lambda x_: x_[1]))


def project_coords(x, y, src_crs, dst_crs, return_as='1d', **kwargs):

    """
    Projects coordinates to a new CRS

    Args:
        x (1d array-like): The x coordinates.
        y (1d array-like): The y coordinates.
        src_crs (str, dict, object): The source CRS.
        dst_crs (str, dict, object): The destination CRS.
        return_as (Optional[str]): How to return the coordinates. Choices are ['1d', '2d'].
        kwargs (Optional[dict]): Keyword arguments passed to ``rasterio.warp.reproject``.

    Returns:
        ``numpy.array``, ``numpy.array`` or ``xr.DataArray``
    """

    if return_as == '1d':

        df_tmp = gpd.GeoDataFrame(np.arange(0, x.shape[0]),
                                  geometry=gpd.points_from_xy(x, y),
                                  crs=src_crs)

        df_tmp = df_tmp.to_crs(dst_crs)

        return df_tmp.geometry.x.values, df_tmp.geometry.y.values

    else:

        if not isinstance(x, np.ndarray):
            x = np.array(x, dtype='float64')

        if not isinstance(y, np.ndarray):
            y = np.array(y, dtype='float64')

        yy = np.meshgrid(x, y)[1]

        latitudes = np.zeros(yy.shape, dtype='float64')

        src_transform = from_bounds(x[0], y[-1], x[-1], y[0], latitudes.shape[1], latitudes.shape[0])

        west, south, east, north = transform_bounds(src_crs, CRS(dst_crs), x[0], y[-1], x[-1], y[0])
        dst_transform = from_bounds(west, south, east, north, latitudes.shape[1], latitudes.shape[0])

        latitudes = reproject(yy,
                              destination=latitudes,
                              src_transform=src_transform,
                              dst_transform=dst_transform,
                              src_crs=CRS.from_epsg(src_crs.split(':')[1]),
                              dst_crs=CRS(dst_crs),
                              **kwargs)[0]

        return xr.DataArray(data=da.from_array(latitudes[np.newaxis, :, :],
                                               chunks=(1, 512, 512)),
                            dims=('band', 'y', 'x'),
                            coords={'band': ['lat'],
                                    'y': ('y', latitudes[:, 0]),
                                    'x': ('x', np.arange(1, latitudes.shape[1]+1))})


def get_geometry_info(geometry, res):

    """
    Gets information from a Shapely geometry object

    Args:
        geometry (object): A `shapely.geometry` object.
        res (tuple): The cell resolution for the affine transform.

    Returns:
        Geometry information (namedtuple)
    """

    GeomInfo = namedtuple('GeomInfo', 'left bottom right top shape affine')

    minx, miny, maxx, maxy = geometry.bounds

    if isinstance(minx, str):
        minx, miny, maxx, maxy = geometry.bounds.values[0]

    out_shape = (int((maxy - miny) / res[1]), int((maxx - minx) / res[0]))

    return GeomInfo(left=minx,
                    bottom=miny,
                    right=maxx,
                    top=maxy,
                    shape=out_shape,
                    affine=Affine(res[0], 0.0, minx, 0.0, -res[1], maxy))


def get_file_extension(filename):

    """
    Gets file and directory name information

    Args:
        filename (str): The file name.

    Returns:
        Name information (namedtuple)
    """

    FileNames = namedtuple('FileNames', 'd_name f_name f_base f_ext')

    d_name, f_name = os.path.splitext(filename)
    f_base, f_ext = os.path.split(f_name)

    return FileNames(d_name=d_name, f_name=f_name, f_base=f_base, f_ext=f_ext)


def n_rows_cols(pixel_index, block_size, rows_cols):

    """
    Adjusts block size for the end of image rows and columns.

    Args:
        pixel_index (int): The current pixel row or column index.
        block_size (int): The image block size.
        rows_cols (int): The total number of rows or columns in the image.

    Returns:
        Adjusted block size as int.
    """

    return block_size if (pixel_index + block_size) < rows_cols else rows_cols - pixel_index


class Chunks(object):

    @staticmethod
    def get_chunk_dim(chunksize):
        return '{:d}d'.format(len(chunksize))

    def check_chunktype(self, chunksize, output='3d'):

        if isinstance(chunksize, int):
            chunksize = (1, chunksize, chunksize)

        chunk_len = len(chunksize)
        output_len = int(output[0])

        if not isinstance(chunksize, tuple):
            if not isinstance(chunksize, dict):
                if not isinstance(chunksize, int):
                    logger.warning('  The chunksize parameter should be a tuple, dictionary, or integer.')

        # TODO: make compatible with multi-layer predictions (e.g., probabilities)
        if chunk_len != output_len:
            self.check_chunksize(chunksize, output=output)

    @staticmethod
    def check_chunksize(chunksize, output='3d'):

        chunk_len = len(chunksize)
        output_len = int(output[0])

        if chunk_len != output_len:

            if (chunk_len == 2) and (output_len == 3):
                return (1,) + chunksize
            elif (chunk_len == 3) and (output_len == 2):
                return chunksize[1:]

        return chunksize


class MapProcesses(object):

    @staticmethod
    def moving(data,
               band_names=None,
               stat='mean',
               perc=50,
               w=3,
               nodata=None,
               n_jobs=1):

        """
        Applies a moving window function over Dask array blocks

        Args:
            data (DataArray): The ``xarray.DataArray`` to process.
            band_names (int or str or list): The output band name(s).
            stat (Optional[str]): The statistic to compute. Choices are ['mean', 'std', 'var', 'min', 'max', 'perc'].
            perc (Optional[int]): The percentile to return if ``stat`` = 'perc'.
            nodata (Optional[int or float]): A 'no data' value to ignore.
            w (Optional[int]): The moving window size (in pixels).
            n_jobs (Optional[int]): The number of rows to process in parallel.

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>>
            >>> # Calculate the mean within a 5x5 window
            >>> with gw.open('image.tif') as src:
            >>>     res = gw.moving(ds, stat='mean', w=5, nodata=32767.0, n_jobs=8)
            >>>
            >>> # Calculate the 90th percentile within a 15x15 window
            >>> with gw.open('image.tif') as src:
            >>>     res = gw.moving(stat='perc', w=15, perc=90, nodata=32767.0, n_jobs=8)
        """

        if w % 2 == 0:
            logger.exception('  The window size must be an odd number.')

        if not isinstance(data, xr.DataArray):
            logger.exception('  The input data must be an Xarray DataArray.')

        y = data.y.values
        x = data.x.values
        attrs = data.attrs

        if n_jobs <= 0:

            logger.warning('  The number of parallel jobs should be a positive integer, so setting n_jobs=1.')
            n_jobs = 1

        hw = int(w / 2.0)

        def _move_func(block_data):

            """
            Args:
                block_data (2d array)
            """

            if max(block_data.shape) <= hw:
                return block_data
            else:

                return moving_window(block_data,
                                     stat=stat,
                                     w=w,
                                     perc=perc,
                                     nodata=nodata,
                                     n_jobs=n_jobs)

        results = list()

        for band in data.band.values.tolist():

            band_array = data.sel(band=band)

            res = band_array.astype('float64').data.map_overlap(_move_func,
                                                                depth=(hw, hw),
                                                                trim=True,
                                                                boundary='reflect',
                                                                dtype='float64')

            results.append(res)

        if isinstance(band_names, np.ndarray):

            if isinstance(band_names.tolist(), str):
                band_names = [band_names.tolist()]

        if not isinstance(band_names, list):

            if not isinstance(band_names, np.ndarray):
                band_names = np.arange(1, data.shape[0]+1)

        results = xr.DataArray(data=da.stack(results, axis=0),
                               dims=('band', 'y', 'x'),
                               coords={'band': band_names,
                                       'y': y,
                                       'x': x},
                               attrs=attrs)

        results.attrs['moving_stat'] = stat
        results.attrs['moving_window_size'] = w

        if stat == 'perc':
            results.attrs['moving_perc'] = perc

        return results


def sample_feature(fid, geom, crs, res, all_touched, meta, frac, feature_array=None):

    """
    Samples a polygon features

    Args:
        fid
        geom
        crs
        res
        all_touched
        meta
        frac
        feature_array

    Returns:
        ``geopandas.GeoDataFrame``
    """

    # Get the feature's bounding extent
    geom_info = get_geometry_info(geom, res)

    if min(geom_info.shape) == 0:
        return gpd.GeoDataFrame([])

    if not isinstance(feature_array, np.ndarray):
    
        # "Rasterize" the geometry into a NumPy array
        feature_array = features.rasterize([geom],
                                           out_shape=geom_info.shape,
                                           fill=0,
                                           out=None,
                                           transform=geom_info.affine,
                                           all_touched=all_touched,
                                           default_value=1,
                                           dtype='int32')

    # Get the indices of the feature's envelope
    valid_samples = np.where(feature_array == 1)

    # Convert the indices to map indices
    y_samples = valid_samples[0] + int(round(abs(meta.top - geom_info.top)) / res[1])
    x_samples = valid_samples[1] + int(round(abs(geom_info.left - meta.left)) / res[0])

    # Convert the map indices to map coordinates
    x_coords, y_coords = meta.affine * (x_samples, y_samples)

    # y_coords = meta.top - y_samples * data.res[0]
    # x_coords = meta.left + x_samples * data.res[0]

    if frac < 1:

        rand_idx = np.random.choice(np.arange(0, y_coords.shape[0]),
                                    size=int(y_coords.shape[0] * frac),
                                    replace=False)

        y_coords = y_coords[rand_idx]
        x_coords = x_coords[rand_idx]

    n_samples = y_coords.shape[0]

    try:

        fid_ = int(fid)
        fid_ = np.zeros(n_samples, dtype='int64') + fid_

    except:

        fid_ = str(fid)
        fid_ = np.zeros([fid_]*n_samples, dtype=object)

    # Combine the coordinates into `Shapely` point geometry
    return gpd.GeoDataFrame(data=np.c_[fid_, np.arange(0, n_samples)],
                            geometry=gpd.points_from_xy(x_coords, y_coords),
                            crs=crs,
                            columns=['poly', 'point'])
